@@ -24,8 +24,11 @@
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
+
 #include "uv.h"
 #include "internal.h"
+#include "handle-inl.h"
+#include "req-inl.h"
 
 
 const unsigned int uv_directory_watcher_buffer_size = 4096;
@@ -33,9 +36,7 @@ const unsigned int uv_directory_watcher_buffer_size = 4096;
 
 static void uv_fs_event_init_handle(uv_loop_t* loop, uv_fs_event_t* handle,
     const char* filename, uv_fs_event_cb cb) {
-  handle->type = UV_FS_EVENT;
-  handle->loop = loop;
-  handle->flags = 0;
+  uv__handle_init(loop, (uv_handle_t*) handle, UV_FS_EVENT);
   handle->cb = cb;
   handle->dir_handle = INVALID_HANDLE_VALUE;
   handle->buffer = NULL;
@@ -53,10 +54,7 @@ static void uv_fs_event_init_handle(uv_loop_t* loop, uv_fs_event_t* handle,
     uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
   }
 
-  loop->counters.handle_init++;
-  loop->counters.fs_event_init++;
-
-  uv_ref(loop);
+  uv__handle_start(handle);
 }
 
 
@@ -90,15 +88,15 @@ static void uv_fs_event_queue_readdirchanges(uv_loop_t* loop,
 }
 
 
-static int uv_split_path(const wchar_t* filename, wchar_t** dir,
-    wchar_t** file) {
+static int uv_split_path(const WCHAR* filename, WCHAR** dir,
+    WCHAR** file) {
   int len = wcslen(filename);
   int i = len;
   while (i > 0 && filename[--i] != '\\' && filename[i] != '/');
 
   if (i == 0) {
     if (dir) {
-      *dir = (wchar_t*)malloc((MAX_PATH + 1) * sizeof(wchar_t));
+      *dir = (WCHAR*)malloc((MAX_PATH + 1) * sizeof(WCHAR));
       if (!*dir) {
         uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
       }
@@ -109,11 +107,11 @@ static int uv_split_path(const wchar_t* filename, wchar_t** dir,
         return -1;
       }
     }
-    
+
     *file = wcsdup(filename);
   } else {
     if (dir) {
-      *dir = (wchar_t*)malloc((i + 1) * sizeof(wchar_t));
+      *dir = (WCHAR*)malloc((i + 1) * sizeof(WCHAR));
       if (!*dir) {
         uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
       }
@@ -121,7 +119,7 @@ static int uv_split_path(const wchar_t* filename, wchar_t** dir,
       (*dir)[i] = L'\0';
     }
 
-    *file = (wchar_t*)malloc((len - i) * sizeof(wchar_t));
+    *file = (WCHAR*)malloc((len - i) * sizeof(WCHAR));
     if (!*file) {
       uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
     }
@@ -137,23 +135,20 @@ int uv_fs_event_init(uv_loop_t* loop, uv_fs_event_t* handle,
     const char* filename, uv_fs_event_cb cb, int flags) {
   int name_size, is_path_dir;
   DWORD attr, last_error;
-  wchar_t* dir = NULL, *dir_to_watch, *filenamew = NULL;
-  wchar_t short_path[MAX_PATH];
-
-  /* We don't support any flags yet. */
-  assert(!flags);
+  WCHAR* dir = NULL, *dir_to_watch, *filenamew = NULL;
+  WCHAR short_path[MAX_PATH];
 
   uv_fs_event_init_handle(loop, handle, filename, cb);
 
   /* Convert name to UTF16. */
-  name_size = uv_utf8_to_utf16(filename, NULL, 0) * sizeof(wchar_t);
-  filenamew = (wchar_t*)malloc(name_size);
+  name_size = uv_utf8_to_utf16(filename, NULL, 0) * sizeof(WCHAR);
+  filenamew = (WCHAR*)malloc(name_size);
   if (!filenamew) {
     uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
   }
 
-  if (!uv_utf8_to_utf16(filename, filenamew, 
-      name_size / sizeof(wchar_t))) {
+  if (!uv_utf8_to_utf16(filename, filenamew,
+      name_size / sizeof(WCHAR))) {
     uv__set_sys_error(loop, GetLastError());
     return -1;
   }
@@ -172,13 +167,13 @@ int uv_fs_event_init(uv_loop_t* loop, uv_fs_event_t* handle,
     handle->dirw = filenamew;
     dir_to_watch = filenamew;
   } else {
-    /* 
+    /*
      * filename is a file.  So we split filename into dir & file parts, and
      * watch the dir directory.
      */
-    
+
     /* Convert to short path. */
-    if (!GetShortPathNameW(filenamew, short_path, COUNTOF(short_path))) {
+    if (!GetShortPathNameW(filenamew, short_path, ARRAY_SIZE(short_path))) {
       last_error = GetLastError();
       goto error;
     }
@@ -226,7 +221,7 @@ int uv_fs_event_init(uv_loop_t* loop, uv_fs_event_t* handle,
     goto error;
   }
 
-  handle->buffer = (char*)_aligned_malloc(uv_directory_watcher_buffer_size, 
+  handle->buffer = (char*)_aligned_malloc(uv_directory_watcher_buffer_size,
     sizeof(DWORD));
   if (!handle->buffer) {
     uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
@@ -294,7 +289,7 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
   FILE_NOTIFY_INFORMATION* file_info;
   int sizew, size, result;
   char* filename = NULL;
-  wchar_t* filenamew, *long_filenamew = NULL;
+  WCHAR* filenamew, *long_filenamew = NULL;
   DWORD offset = 0;
 
   assert(req->type == UV_FS_EVENT_REQ);
@@ -303,7 +298,7 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
 
   /* If we're closing, don't report any callbacks, and just push the handle */
   /* onto the endgame queue. */
-  if (handle->flags & UV_HANDLE_CLOSING) {
+  if (handle->flags & UV__HANDLE_CLOSING) {
     uv_want_endgame(loop, (uv_handle_t*) handle);
     return;
   };
@@ -317,19 +312,19 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
         assert(!filename);
         assert(!long_filenamew);
 
-        /* 
+        /*
          * Fire the event only if we were asked to watch a directory,
          * or if the filename filter matches.
          */
         if (handle->dirw ||
           _wcsnicmp(handle->filew, file_info->FileName,
-            file_info->FileNameLength / sizeof(wchar_t)) == 0 ||
+            file_info->FileNameLength / sizeof(WCHAR)) == 0 ||
           _wcsnicmp(handle->short_filew, file_info->FileName,
-            file_info->FileNameLength / sizeof(wchar_t)) == 0) {
+            file_info->FileNameLength / sizeof(WCHAR)) == 0) {
 
           if (handle->dirw) {
-            /* 
-             * We attempt to convert the file name to its long form for 
+            /*
+             * We attempt to convert the file name to its long form for
              * events that still point to valid files on disk.
              * For removed and renamed events, we do not provide the file name.
              */
@@ -337,14 +332,15 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
               file_info->Action != FILE_ACTION_RENAMED_OLD_NAME) {
               /* Construct a full path to the file. */
               size = wcslen(handle->dirw) +
-                file_info->FileNameLength / sizeof(wchar_t) + 2;
+                file_info->FileNameLength / sizeof(WCHAR) + 2;
 
-              filenamew = (wchar_t*)malloc(size * sizeof(wchar_t));
+              filenamew = (WCHAR*)malloc(size * sizeof(WCHAR));
               if (!filenamew) {
                 uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
               }
 
-              _snwprintf(filenamew, size, L"%s\\%s", handle->dirw,
+              _snwprintf(filenamew, size, L"%s\\%.*s", handle->dirw,
+                file_info->FileNameLength / sizeof(WCHAR),
                 file_info->FileName);
 
               filenamew[size - 1] = L'\0';
@@ -353,7 +349,7 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
               size = GetLongPathNameW(filenamew, NULL, 0);
 
               if (size) {
-                long_filenamew = (wchar_t*)malloc(size * sizeof(wchar_t));
+                long_filenamew = (WCHAR*)malloc(size * sizeof(WCHAR));
                 if (!long_filenamew) {
                   uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
                 }
@@ -382,13 +378,13 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
                 }
               }
 
-              /* 
+              /*
                * If we couldn't get the long name - just use the name
                * provided by ReadDirectoryChangesW.
                */
               if (!long_filenamew) {
                 filenamew = file_info->FileName;
-                sizew = file_info->FileNameLength / sizeof(wchar_t);
+                sizew = file_info->FileNameLength / sizeof(WCHAR);
               }
             } else {
               /* Removed or renamed callbacks don't provide filename. */
@@ -445,7 +441,7 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
         }
 
         offset = file_info->NextEntryOffset;
-      } while (offset && !(handle->flags & UV_HANDLE_CLOSING));
+      } while (offset && !(handle->flags & UV__HANDLE_CLOSING));
     } else {
       handle->cb(handle, NULL, UV_CHANGE, 0);
     }
@@ -454,7 +450,7 @@ void uv_process_fs_event_req(uv_loop_t* loop, uv_req_t* req,
     handle->cb(handle, NULL, 0, -1);
   }
 
-  if (!(handle->flags & UV_HANDLE_CLOSING)) {
+  if (!(handle->flags & UV__HANDLE_CLOSING)) {
     uv_fs_event_queue_readdirchanges(loop, handle);
   } else {
     uv_want_endgame(loop, (uv_handle_t*)handle);
@@ -471,14 +467,15 @@ void uv_fs_event_close(uv_loop_t* loop, uv_fs_event_t* handle) {
   if (!handle->req_pending) {
     uv_want_endgame(loop, (uv_handle_t*)handle);
   }
+
+  uv__handle_closing(handle);
 }
 
 
 void uv_fs_event_endgame(uv_loop_t* loop, uv_fs_event_t* handle) {
-  if (handle->flags & UV_HANDLE_CLOSING &&
+  if (handle->flags & UV__HANDLE_CLOSING &&
       !handle->req_pending) {
     assert(!(handle->flags & UV_HANDLE_CLOSED));
-    handle->flags |= UV_HANDLE_CLOSED;
 
     if (handle->buffer) {
       _aligned_free(handle->buffer);
@@ -505,10 +502,6 @@ void uv_fs_event_endgame(uv_loop_t* loop, uv_fs_event_t* handle) {
       handle->dirw = NULL;
     }
 
-    if (handle->close_cb) {
-      handle->close_cb((uv_handle_t*)handle);
-    }
-
-    uv_unref(loop);
+    uv__handle_close(handle);
   }
 }
