@@ -28,36 +28,20 @@ function errnoException(errorno, syscall) {
   // TODO make this more compatible with ErrnoException from src/node.cc
   // Once all of Node is using this function the ErrnoException from
   // src/node.cc should be removed.
-  var e = new Error(syscall + ' ' + errorno);
 
   // For backwards compatibility. libuv returns ENOENT on NXDOMAIN.
   if (errorno == 'ENOENT') {
-    errorno = 'ENOTFOUND'
+    errorno = 'ENOTFOUND';
   }
 
+  var e = new Error(syscall + ' ' + errorno);
+
   e.errno = e.code = errorno;
   e.syscall = syscall;
   return e;
 }
 
 
-function familyToSym(family) {
-  switch (family) {
-    case 4: return cares.AF_INET;
-    case 6: return cares.AF_INET6;
-    default: return cares.AF_UNSPEC;
-  }
-}
-
-
-function symToFamily(family) {
-  switch (family) {
-    case cares.AF_INET: return 4;
-    case cares.AF_INET6: return 6;
-    default: return undefined;
-  }
-}
-
 // c-ares invokes a callback either synchronously or asynchronously,
 // but the dns API should always invoke a callback asynchronously.
 //
@@ -137,14 +121,14 @@ exports.lookup = function(domain, family, callback) {
         callback(null, addresses[0], addresses[0].indexOf(':') >= 0 ? 6 : 4);
       }
     } else {
-      callback(errnoException(errno, 'getaddrinfo'));
+      callback(errnoException(process._errno, 'getaddrinfo'));
     }
   }
 
   var wrap = cares.getaddrinfo(domain, family);
 
   if (!wrap) {
-    throw errnoException(errno, 'getaddrinfo');
+    throw errnoException(process._errno, 'getaddrinfo');
   }
 
   wrap.oncomplete = onanswer;
@@ -162,14 +146,14 @@ function resolver(bindingName) {
       if (!status) {
         callback(null, result);
       } else {
-        callback(errnoException(errno, bindingName));
+        callback(errnoException(process._errno, bindingName));
       }
     }
 
     callback = makeAsync(callback);
     var wrap = binding(name, onanswer);
     if (!wrap) {
-      throw errnoException(errno, bindingName);
+      throw errnoException(process._errno, bindingName);
     }
 
     callback.immediately = true;
@@ -186,6 +170,7 @@ exports.resolveMx = resolveMap.MX = resolver('queryMx');
 exports.resolveNs = resolveMap.NS = resolver('queryNs');
 exports.resolveTxt = resolveMap.TXT = resolver('queryTxt');
 exports.resolveSrv = resolveMap.SRV = resolver('querySrv');
+exports.resolveNaptr = resolveMap.NAPTR = resolver('queryNaptr');
 exports.reverse = resolveMap.PTR = resolver('getHostByAddr');
 
 
@@ -202,21 +187,33 @@ exports.resolve = function(domain, type_, callback_) {
   if (typeof resolver === 'function') {
     return resolver(domain, callback);
   } else {
-    throw new Error('Unknown type "' + type + '"');
+    throw new Error('Unknown type "' + type_ + '"');
   }
 };
 
 
 // ERROR CODES
-exports.BADNAME = 'EBADNAME';
-exports.BADRESP = 'EBADRESP';
-exports.CONNREFUSED = 'ECONNREFUSED';
-exports.DESTRUCTION = 'EDESTRUCTION';
-exports.REFUSED = 'EREFUSED';
-exports.FORMERR = 'EFORMERR';
 exports.NODATA = 'ENODATA';
-exports.NOMEM = 'ENOMEM';
+exports.FORMERR = 'EFORMERR';
+exports.SERVFAIL = 'ESERVFAIL';
 exports.NOTFOUND = 'ENOTFOUND';
 exports.NOTIMP = 'ENOTIMP';
-exports.SERVFAIL = 'ESERVFAIL';
+exports.REFUSED = 'EREFUSED';
+exports.BADQUERY = 'EBADQUERY';
+exports.ADNAME = 'EADNAME';
+exports.BADFAMILY = 'EBADFAMILY';
+exports.BADRESP = 'EBADRESP';
+exports.CONNREFUSED = 'ECONNREFUSED';
 exports.TIMEOUT = 'ETIMEOUT';
+exports.EOF = 'EOF';
+exports.FILE = 'EFILE';
+exports.NOMEM = 'ENOMEM';
+exports.DESTRUCTION = 'EDESTRUCTION';
+exports.BADSTR = 'EBADSTR';
+exports.BADFLAGS = 'EBADFLAGS';
+exports.NONAME = 'ENONAME';
+exports.BADHINTS = 'EBADHINTS';
+exports.NOTINITIALIZED = 'ENOTINITIALIZED';
+exports.LOADIPHLPAPI = 'ELOADIPHLPAPI';
+exports.ADDRGETNETWORKPARAMS = 'EADDRGETNETWORKPARAMS';
+exports.CANCELLED = 'ECANCELLED';
