@@ -45,7 +45,7 @@ exports.EventEmitter = EventEmitter;
 // that to be increased. Set to zero for unlimited.
 var defaultMaxListeners = 10;
 EventEmitter.prototype.setMaxListeners = function(n) {
-  if (typeof n !== 'number' || n < 0)
+  if (typeof n !== 'number' || n < 0 || isNaN(n))
     throw TypeError('n must be a positive number');
   this._maxListeners = n;
 };
@@ -170,9 +170,15 @@ EventEmitter.prototype.once = function(type, listener) {
   if (typeof listener !== 'function')
     throw TypeError('listener must be a function');
 
+  var fired = false;
+
   function g() {
     this.removeListener(type, g);
-    listener.apply(this, arguments);
+
+    if (!fired) {
+      fired = true;
+      listener.apply(this, arguments);
+    }
   }
 
   g.listener = listener;
@@ -197,7 +203,7 @@ EventEmitter.prototype.removeListener = function(type, listener) {
 
   if (list === listener ||
       (typeof list.listener === 'function' && list.listener === listener)) {
-    this._events[type] = undefined;
+    delete this._events[type];
     if (this._events.removeListener)
       this.emit('removeListener', type, listener);
 
@@ -215,7 +221,7 @@ EventEmitter.prototype.removeListener = function(type, listener) {
 
     if (list.length === 1) {
       list.length = 0;
-      this._events[type] = undefined;
+      delete this._events[type];
     } else {
       list.splice(position, 1);
     }
@@ -238,7 +244,7 @@ EventEmitter.prototype.removeAllListeners = function(type) {
     if (arguments.length === 0)
       this._events = {};
     else if (this._events[type])
-      this._events[type] = undefined;
+      delete this._events[type];
     return this;
   }
 
@@ -257,12 +263,12 @@ EventEmitter.prototype.removeAllListeners = function(type) {
 
   if (typeof listeners === 'function') {
     this.removeListener(type, listeners);
-  } else {
+  } else if (Array.isArray(listeners)) {
     // LIFO order
     while (listeners.length)
       this.removeListener(type, listeners[listeners.length - 1]);
   }
-  this._events[type] = undefined;
+  delete this._events[type];
 
   return this;
 };
