@@ -21,6 +21,8 @@
 
 var tls = require('tls');
 var http = require('http');
+var util = require('util');
+var url = require('url');
 var inherits = require('util').inherits;
 
 function Server(opts, requestListener) {
@@ -37,6 +39,10 @@ function Server(opts, requestListener) {
   if (requestListener) {
     this.addListener('request', requestListener);
   }
+
+  this.addListener('clientError', function(err, conn) {
+    conn.destroy(err);
+  });
 }
 inherits(Server, tls.Server);
 
@@ -52,13 +58,32 @@ exports.createServer = function(opts, requestListener) {
 // HTTPS agents.
 
 function createConnection(port, host, options) {
-  return tls.connect(port, host, options);
-};
+  if (typeof port === 'object') {
+    options = port;
+  } else if (typeof host === 'object') {
+    options = host;
+  } else if (typeof options === 'object') {
+    options = options;
+  } else {
+    options = {};
+  }
+
+  if (typeof port === 'number') {
+    options.port = port;
+  }
+
+  if (typeof host === 'string') {
+    options.host = host;
+  }
+
+  return tls.connect(options);
+}
+
 
 function Agent(options) {
   http.Agent.call(this, options);
   this.createConnection = createConnection;
-};
+}
 inherits(Agent, http.Agent);
 Agent.prototype.defaultPort = 443;
 
@@ -68,20 +93,37 @@ exports.globalAgent = globalAgent;
 exports.Agent = Agent;
 
 exports.request = function(options, cb) {
+  if (typeof options === 'string') {
+    options = url.parse(options);
+  }
+
   if (options.protocol && options.protocol !== 'https:') {
     throw new Error('Protocol:' + options.protocol + ' not supported.');
   }
 
-  if (options.agent === undefined) {
-    options.agent = globalAgent;
+  options = util._extend({
+    createConnection: createConnection,
+    defaultPort: 443
+  }, options);
+
+  if (typeof options.agent === 'undefined') {
+    if (typeof options.ca === 'undefined' &&
+        typeof options.cert === 'undefined' &&
+        typeof options.ciphers === 'undefined' &&
+        typeof options.key === 'undefined' &&
+        typeof options.passphrase === 'undefined' &&
+        typeof options.pfx === 'undefined' &&
+        typeof options.rejectUnauthorized === 'undefined') {
+      options.agent = globalAgent;
+    } else {
+      options.agent = new Agent(options);
+    }
   }
-  options.createConnection = createConnection;
-  options.defaultPort = options.defaultPort || 443;
+
   return new http.ClientRequest(options, cb);
 };
 
 exports.get = function(options, cb) {
-  options.method = 'GET';
   var req = exports.request(options, cb);
   req.end();
   return req;
