@@ -20,10 +20,11 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 var events = require('events');
-var Stream = require('stream');
+var stream = require('stream');
 var timers = require('timers');
 var util = require('util');
 var assert = require('assert');
+var cares = process.binding('cares_wrap');
 var cluster;
 
 function noop() {}
@@ -41,16 +42,25 @@ function createTCP() {
 }
 
 
-/* Bit flags for socket._flags */
-var FLAG_GOT_EOF = 1 << 0;
-var FLAG_SHUTDOWN = 1 << 1;
-var FLAG_DESTROY_SOON = 1 << 2;
-var FLAG_SHUTDOWN_QUEUED = 1 << 3;
+function createHandle(fd) {
+  var tty = process.binding('tty_wrap');
+  var type = tty.guessHandleType(fd);
+  if (type === 'PIPE') return createPipe();
+  if (type === 'TCP') return createTCP();
+  throw new TypeError('Unsupported fd type: ' + type);
+}
 
 
 var debug;
 if (process.env.NODE_DEBUG && /net/.test(process.env.NODE_DEBUG)) {
-  debug = function(x) { console.error('NET:', x); };
+  var pid = process.pid;
+  debug = function(x) {
+    // if console is not set up yet, then skip this.
+    if (!console.error)
+      return;
+    console.error('NET: %d', pid,
+                  util.format.apply(util, arguments).slice(0, 500));
+  };
 } else {
   debug = function() { };
 }
@@ -109,12 +119,8 @@ function normalizeConnectArgs(args) {
 exports._normalizeConnectArgs = normalizeConnectArgs;
 
 
-/* called when creating new Socket, or when re-using a closed Socket */
+// called when creating new Socket, or when re-using a closed Socket
 function initSocketHandle(self) {
-  self._pendingWriteReqs = 0;
-
-  self._flags = 0;
-  self._connectQueueSize = 0;
   self.destroyed = false;
   self.errorEmitted = false;
   self.bytesRead = 0;
@@ -130,7 +136,8 @@ function initSocketHandle(self) {
 function Socket(options) {
   if (!(this instanceof Socket)) return new Socket(options);
 
-  Stream.call(this);
+  this._connecting = false;
+  this._handle = null;
 
   switch (typeof options) {
     case 'number':
@@ -141,25 +148,158 @@ function Socket(options) {
       break;
   }
 
-  if (typeof options.fd === 'undefined') {
-    this._handle = options && options.handle; // private
-  } else {
-    this._handle = createPipe();
+  stream.Duplex.call(this, options);
+
+  if (options.handle) {
+    this._handle = options.handle; // private
+  } else if (typeof options.fd !== 'undefined') {
+    this._handle = createHandle(options.fd);
     this._handle.open(options.fd);
-    this.readable = this.writable = true;
+    this.readable = options.readable !== false;
+    this.writable = options.writable !== false;
+  } else {
+    // these will be set once there is a connection
+    this.readable = this.writable = false;
   }
 
-  initSocketHandle(this);
-  this.allowHalfOpen = options && options.allowHalfOpen;
-}
-util.inherits(Socket, Stream);
+  this.onend = null;
 
+  // shut down the socket when we're finished with it.
+  this.on('finish', onSocketFinish);
+  this.on('_socketEnd', onSocketEnd);
+
+  initSocketHandle(this);
+
+  this._pendingData = null;
+  this._pendingEncoding = '';
+
+  // handle strings directly
+  this._writableState.decodeStrings = false;
+
+  // default to *not* allowing half open sockets
+  this.allowHalfOpen = options && options.allowHalfOpen || false;
+
+  // if we have a handle, then start the flow of data into the
+  // buffer.  if not, then this will happen when we connect
+  if (this._handle && options.readable !== false)
+    this.read(0);
+}
+util.inherits(Socket, stream.Duplex);
+
+// the user has called .end(), and all the bytes have been
+// sent out to the other side.
+// If allowHalfOpen is false, or if the readable side has
+// ended already, then destroy.
+// If allowHalfOpen is true, then we need to do a shutdown,
+// so that only the writable side will be cleaned up.
+function onSocketFinish() {
+  // If still connecting - defer handling 'finish' until 'connect' will happen
+  if (this._connecting) {
+    debug('osF: not yet connected');
+    return this.once('connect', onSocketFinish);
+  }
+
+  debug('onSocketFinish');
+  if (!this.readable || this._readableState.ended) {
+    debug('oSF: ended, destroy', this._readableState);
+    return this.destroy();
+  }
+
+  debug('oSF: not ended, call shutdown()');
+
+  // otherwise, just shutdown, or destroy() if not possible
+  if (!this._handle || !this._handle.shutdown)
+    return this.destroy();
+
+  var shutdownReq = this._handle.shutdown();
+
+  if (!shutdownReq)
+    return this._destroy(errnoException(process._errno, 'shutdown'));
+
+  shutdownReq.oncomplete = afterShutdown;
+}
+
+
+function afterShutdown(status, handle, req) {
+  var self = handle.owner;
+
+  debug('afterShutdown destroyed=%j', self.destroyed,
+        self._readableState);
+
+  // callback may come after call to destroy.
+  if (self.destroyed)
+    return;
+
+  if (self._readableState.ended) {
+    debug('readableState ended, destroying');
+    self.destroy();
+  } else {
+    self.once('_socketEnd', self.destroy);
+  }
+}
+
+// the EOF has been received, and no more bytes are coming.
+// if the writable side has ended already, then clean everything
+// up.
+function onSocketEnd() {
+  // XXX Should not have to do as much crap in this function.
+  // ended should already be true, since this is called *after*
+  // the EOF errno and onread has eof'ed
+  debug('onSocketEnd', this._readableState);
+  this._readableState.ended = true;
+  if (this._readableState.endEmitted) {
+    this.readable = false;
+    maybeDestroy(this);
+  } else {
+    this.once('end', function() {
+      this.readable = false;
+      maybeDestroy(this);
+    });
+    this.read(0);
+  }
+
+  if (!this.allowHalfOpen) {
+    this.write = writeAfterFIN;
+    this.destroySoon();
+  }
+}
+
+// Provide a better error message when we call end() as a result
+// of the other side sending a FIN.  The standard 'write after end'
+// is overly vague, and makes it seem like the user's code is to blame.
+function writeAfterFIN(chunk, encoding, cb) {
+  if (typeof encoding === 'function') {
+    cb = encoding;
+    encoding = null;
+  }
+
+  var er = new Error('This socket has been ended by the other party');
+  er.code = 'EPIPE';
+  var self = this;
+  // TODO: defer error events consistently everywhere, not just the cb
+  self.emit('error', er);
+  if (typeof cb === 'function') {
+    process.nextTick(function() {
+      cb(er);
+    });
+  }
+}
 
 exports.Socket = Socket;
 exports.Stream = Socket; // Legacy naming.
 
+Socket.prototype.read = function(n) {
+  if (n === 0)
+    return stream.Readable.prototype.read.call(this, n);
+
+  this.read = stream.Readable.prototype.read;
+  this._consuming = true;
+  return this.read(n);
+};
+
 
 Socket.prototype.listen = function() {
+  debug('socket.listen');
   var self = this;
   self.on('connection', arguments[0]);
   listen(self, null, null, null);
@@ -169,7 +309,7 @@ Socket.prototype.listen = function() {
 Socket.prototype.setTimeout = function(msecs, callback) {
   if (msecs > 0 && !isNaN(msecs) && isFinite(msecs)) {
     timers.enroll(this, msecs);
-    timers.active(this);
+    timers._unrefActive(this);
     if (callback) {
       this.once('timeout', callback);
     }
@@ -229,96 +369,69 @@ Object.defineProperty(Socket.prototype, 'readyState', {
 Object.defineProperty(Socket.prototype, 'bufferSize', {
   get: function() {
     if (this._handle) {
-      return this._handle.writeQueueSize + this._connectQueueSize;
+      return this._handle.writeQueueSize + this._writableState.length;
     }
   }
 });
 
 
-Socket.prototype.pause = function() {
-  this._paused = true;
-  if (this._handle && !this._connecting) {
-    this._handle.readStop();
-  }
-};
+// Just call handle.readStart until we have enough in the buffer
+Socket.prototype._read = function(n) {
+  debug('_read');
 
-
-Socket.prototype.resume = function() {
-  this._paused = false;
-  if (this._handle && !this._connecting) {
-    this._handle.readStart();
+  if (this._connecting || !this._handle) {
+    debug('_read wait for connection');
+    this.once('connect', this._read.bind(this, n));
+  } else if (!this._handle.reading) {
+    // not already reading, start the flow
+    debug('Socket._read readStart');
+    this._handle.reading = true;
+    var r = this._handle.readStart();
+    if (r)
+      this._destroy(errnoException(process._errno, 'read'));
   }
 };
 
 
 Socket.prototype.end = function(data, encoding) {
-  if (this._connecting && ((this._flags & FLAG_SHUTDOWN_QUEUED) == 0)) {
-    // still connecting, add data to buffer
-    if (data) this.write(data, encoding);
-    this.writable = false;
-    this._flags |= FLAG_SHUTDOWN_QUEUED;
-  }
-
-  if (!this.writable) return;
+  stream.Duplex.prototype.end.call(this, data, encoding);
   this.writable = false;
-
-  if (data) this.write(data, encoding);
   DTRACE_NET_STREAM_END(this);
 
-  if (!this.readable) {
-    this.destroySoon();
-  } else {
-    this._flags |= FLAG_SHUTDOWN;
-    var shutdownReq = this._handle.shutdown();
-
-    if (!shutdownReq) {
-      this._destroy(errnoException(errno, 'shutdown'));
-      return false;
-    }
-
-    shutdownReq.oncomplete = afterShutdown;
-  }
-
-  return true;
+  // just in case we're waiting for an EOF.
+  if (this.readable && !this._readableState.endEmitted)
+    this.read(0);
+  else
+    maybeDestroy(this);
 };
 
 
-function afterShutdown(status, handle, req) {
-  var self = handle.owner;
-
-  assert.ok(self._flags & FLAG_SHUTDOWN);
-  assert.ok(!self.writable);
-
-  // callback may come after call to destroy.
-  if (self.destroyed) {
-    return;
-  }
-
-  if (self._flags & FLAG_GOT_EOF || !self.readable) {
-    self._destroy();
-  } else {
+// Call whenever we set writable=false or readable=false
+function maybeDestroy(socket) {
+  if (!socket.readable &&
+      !socket.writable &&
+      !socket.destroyed &&
+      !socket._connecting &&
+      !socket._writableState.length) {
+    socket.destroy();
   }
 }
 
 
 Socket.prototype.destroySoon = function() {
-  this.writable = false;
-  this._flags |= FLAG_DESTROY_SOON;
+  if (this.writable)
+    this.end();
 
-  if (this._pendingWriteReqs == 0) {
-    this._destroy();
-  }
-};
-
-
-Socket.prototype._connectQueueCleanUp = function(exception) {
-  this._connecting = false;
-  this._connectQueueSize = 0;
-  this._connectQueue = null;
+  if (this._writableState.finished)
+    this.destroy();
+  else
+    this.once('finish', this.destroy);
 };
 
 
 Socket.prototype._destroy = function(exception, cb) {
+  debug('destroy');
+
   var self = this;
 
   function fireErrorCallbacks() {
@@ -332,13 +445,12 @@ Socket.prototype._destroy = function(exception, cb) {
   };
 
   if (this.destroyed) {
+    debug('already destroyed, fire error callbacks');
     fireErrorCallbacks();
     return;
   }
 
-  self._connectQueueCleanUp();
-
-  debug('destroy');
+  self._connecting = false;
 
   this.readable = this.writable = false;
 
@@ -346,20 +458,23 @@ Socket.prototype._destroy = function(exception, cb) {
 
   debug('close');
   if (this._handle) {
-    this._handle.close();
+    if (this !== process.stderr)
+      debug('close handle');
+    var isException = exception ? true : false;
+    this._handle.close(function() {
+      debug('emit close');
+      self.emit('close', isException);
+    });
     this._handle.onread = noop;
     this._handle = null;
   }
 
   fireErrorCallbacks();
-
-  process.nextTick(function() {
-    self.emit('close', exception ? true : false);
-  });
-
   this.destroyed = true;
 
   if (this.server) {
+    COUNTER_NET_SERVER_CONNECTION_CLOSE(this);
+    debug('has server');
     this.server._connections--;
     if (this.server._emitCloseIfDrained) {
       this.server._emitCloseIfDrained();
@@ -369,70 +484,80 @@ Socket.prototype._destroy = function(exception, cb) {
 
 
 Socket.prototype.destroy = function(exception) {
+  debug('destroy', exception);
   this._destroy(exception);
 };
 
 
+// This function is called whenever the handle gets a
+// buffer, or when there's an error reading.
 function onread(buffer, offset, length) {
   var handle = this;
   var self = handle.owner;
   assert(handle === self._handle, 'handle != self._handle');
 
-  timers.active(self);
+  timers._unrefActive(self);
 
   var end = offset + length;
+  debug('onread', process._errno, offset, length, end);
 
   if (buffer) {
-    // Emit 'data' event.
+    debug('got data');
 
-    if (self._decoder) {
-      // Emit a string.
-      var string = self._decoder.write(buffer.slice(offset, end));
-      if (string.length) self.emit('data', string);
-    } else {
-      // Emit a slice. Attempt to avoid slicing the buffer if no one is
-      // listening for 'data'.
-      if (self._events && self._events['data']) {
-        self.emit('data', buffer.slice(offset, end));
-      }
+    // read success.
+    // In theory (and in practice) calling readStop right now
+    // will prevent this from being called again until _read() gets
+    // called again.
+
+    // if we didn't get any bytes, that doesn't necessarily mean EOF.
+    // wait for the next one.
+    if (offset === end) {
+      debug('not any data, keep waiting');
+      return;
     }
 
+    // if it's not enough data, we'll just call handle.readStart()
+    // again right away.
     self.bytesRead += length;
 
     // Optimization: emit the original buffer with end points
+    var ret = true;
     if (self.ondata) self.ondata(buffer, offset, end);
+    else ret = self.push(buffer.slice(offset, end));
 
-  } else if (errno == 'EOF') {
-    // EOF
-    self.readable = false;
-
-    assert.ok(!(self._flags & FLAG_GOT_EOF));
-    self._flags |= FLAG_GOT_EOF;
-
-    // We call destroy() before end(). 'close' not emitted until nextTick so
-    // the 'end' event will come first as required.
-    if (!self.writable) self._destroy();
-
-    if (!self.allowHalfOpen) self.end();
-    if (self._events && self._events['end']) self.emit('end');
-    if (self.onend) self.onend();
-  } else {
-    // Error
-    if (errno == 'ECONNRESET') {
-      self._destroy();
-    } else {
-      self._destroy(errnoException(errno, 'read'));
+    if (handle.reading && !ret) {
+      handle.reading = false;
+      debug('readStop');
+      var r = handle.readStop();
+      if (r)
+        self._destroy(errnoException(process._errno, 'read'));
     }
+
+  } else if (process._errno == 'EOF') {
+    debug('EOF');
+
+    if (self._readableState.length === 0) {
+      self.readable = false;
+      maybeDestroy(self);
+    }
+
+    if (self.onend) self.once('end', self.onend);
+
+    // push a null to signal the end of data.
+    self.push(null);
+
+    // internal end event so that we know that the actual socket
+    // is no longer readable, and we can start the shutdown
+    // procedure. No need to wait for all the data to be consumed.
+    self.emit('_socketEnd');
+  } else {
+    debug('error', process._errno);
+    // Error
+    self._destroy(errnoException(process._errno, 'read'));
   }
 }
 
 
-Socket.prototype.setEncoding = function(encoding) {
-  var StringDecoder = require('string_decoder').StringDecoder; // lazy load
-  this._decoder = new StringDecoder(encoding);
-};
-
-
 Socket.prototype._getpeername = function() {
   if (!this._handle || !this._handle.getpeername) {
     return {};
@@ -458,123 +583,118 @@ Socket.prototype.__defineGetter__('remotePort', function() {
 });
 
 
-/*
- * Arguments data, [encoding], [cb]
- */
-Socket.prototype.write = function(data, arg1, arg2) {
-  var encoding, cb;
-
-  // parse arguments
-  if (arg1) {
-    if (typeof arg1 === 'string') {
-      encoding = arg1;
-      cb = arg2;
-    } else if (typeof arg1 === 'function') {
-      cb = arg1;
-    } else {
-      throw new Error('bad arg');
+Socket.prototype._getsockname = function() {
+  if (!this._handle || !this._handle.getsockname) {
+    return {};
+  }
+  if (!this._sockname) {
+    this._sockname = this._handle.getsockname();
+    if (this._sockname === null) {
+      return {};
     }
   }
+  return this._sockname;
+};
 
-  if (typeof data === 'string') {
-    encoding = (encoding || 'utf8').toLowerCase();
-    switch (encoding) {
-      case 'utf8':
-      case 'utf-8':
-      case 'ascii':
-      case 'ucs2':
-      case 'ucs-2':
-      case 'utf16le':
-      case 'utf-16le':
-        // This encoding can be handled in the binding layer.
-        break;
 
-      default:
-        data = new Buffer(data, encoding);
-    }
-  } else if (!Buffer.isBuffer(data)) {
-    throw new TypeError('First argument must be a buffer or a string.');
-  }
+Socket.prototype.__defineGetter__('localAddress', function() {
+  return this._getsockname().address;
+});
 
-  // If we are still connecting, then buffer this for later.
-  if (this._connecting) {
-    this._connectQueueSize += data.length;
-    if (this._connectQueue) {
-      this._connectQueue.push([data, encoding, cb]);
-    } else {
-      this._connectQueue = [[data, encoding, cb]];
-    }
-    return false;
-  }
 
-  return this._write(data, encoding, cb);
+Socket.prototype.__defineGetter__('localPort', function() {
+  return this._getsockname().port;
+});
+
+
+Socket.prototype.write = function(chunk, encoding, cb) {
+  if (typeof chunk !== 'string' && !Buffer.isBuffer(chunk))
+    throw new TypeError('invalid data');
+  return stream.Duplex.prototype.write.apply(this, arguments);
 };
 
 
 Socket.prototype._write = function(data, encoding, cb) {
-  timers.active(this);
+  // If we are still connecting, then buffer this for later.
+  // The Writable logic will buffer up any more writes while
+  // waiting for this one to be done.
+  if (this._connecting) {
+    this._pendingData = data;
+    this._pendingEncoding = encoding;
+    this.once('connect', function() {
+      this._write(data, encoding, cb);
+    });
+    return;
+  }
+  this._pendingData = null;
+  this._pendingEncoding = '';
+
+  timers._unrefActive(this);
 
   if (!this._handle) {
     this._destroy(new Error('This socket is closed.'), cb);
     return false;
   }
 
-  var writeReq;
+  var enc = Buffer.isBuffer(data) ? 'buffer' : encoding;
+  var writeReq = createWriteReq(this._handle, data, enc);
 
-  if (Buffer.isBuffer(data)) {
-    writeReq = this._handle.writeBuffer(data);
-
-  } else {
-    switch (encoding) {
-      case 'utf8':
-      case 'utf-8':
-        writeReq = this._handle.writeUtf8String(data);
-        break;
-
-      case 'ascii':
-        writeReq = this._handle.writeAsciiString(data);
-        break;
-
-      case 'ucs2':
-      case 'ucs-2':
-      case 'utf16le':
-      case 'utf-16le':
-        writeReq = this._handle.writeUcs2String(data);
-        break;
-
-      default:
-        assert(0);
-    }
-  }
-
-  if (!writeReq || typeof writeReq !== 'object') {
-    this._destroy(errnoException(errno, 'write'), cb);
-    return false;
-  }
+  if (!writeReq || typeof writeReq !== 'object')
+    return this._destroy(errnoException(process._errno, 'write'), cb);
 
   writeReq.oncomplete = afterWrite;
-  writeReq.cb = cb;
-
-  this._pendingWriteReqs++;
   this._bytesDispatched += writeReq.bytes;
 
-  return this._handle.writeQueueSize == 0;
+  // If it was entirely flushed, we can write some more right now.
+  // However, if more is left in the queue, then wait until that clears.
+  if (this._handle.writeQueueSize === 0)
+    cb();
+  else
+    writeReq.cb = cb;
 };
 
+function createWriteReq(handle, data, encoding) {
+  switch (encoding) {
+    case 'buffer':
+      return handle.writeBuffer(data);
+
+    case 'utf8':
+    case 'utf-8':
+      return handle.writeUtf8String(data);
+
+    case 'ascii':
+      return handle.writeAsciiString(data);
+
+    case 'ucs2':
+    case 'ucs-2':
+    case 'utf16le':
+    case 'utf-16le':
+      return handle.writeUcs2String(data);
+
+    default:
+      return handle.writeBuffer(new Buffer(data, encoding));
+  }
+}
+
 
 Socket.prototype.__defineGetter__('bytesWritten', function() {
   var bytes = this._bytesDispatched,
-      connectQueue = this._connectQueue;
+      state = this._writableState,
+      data = this._pendingData,
+      encoding = this._pendingEncoding;
 
-  if (connectQueue) {
-    connectQueue.forEach(function(el) {
-      var data = el[0];
-      if (Buffer.isBuffer(data)) {
-        bytes += data.length;
-      } else {
-        bytes += Buffer.byteLength(data, el[1]);
-      }
-    }, this);
+  state.buffer.forEach(function(el) {
+    if (Buffer.isBuffer(el.chunk))
+      bytes += el.chunk.length;
+    else
+      bytes += Buffer.byteLength(el.chunk, el.encoding);
+  });
+
+  if (data) {
+    if (Buffer.isBuffer(data))
+      bytes += data.length;
+    else
+      bytes += Buffer.byteLength(data, encoding);
   }
 
   return bytes;
@@ -583,30 +703,29 @@ Socket.prototype.__defineGetter__('bytesWritten', function() {
 
 function afterWrite(status, handle, req) {
   var self = handle.owner;
+  var state = self._writableState;
+  if (self !== process.stderr && self !== process.stdout)
+    debug('afterWrite', status, req);
 
   // callback may come after call to destroy.
   if (self.destroyed) {
+    debug('afterWrite destroyed');
     return;
   }
 
   if (status) {
-    self._destroy(errnoException(errno, 'write'), req.cb);
+    debug('write failure', errnoException(process._errno, 'write'));
+    self._destroy(errnoException(process._errno, 'write'), req.cb);
     return;
   }
 
-  timers.active(self);
+  timers._unrefActive(self);
 
-  self._pendingWriteReqs--;
+  if (self !== process.stderr && self !== process.stdout)
+    debug('afterWrite call cb');
 
-  if (self._pendingWriteReqs == 0) {
-    self.emit('drain');
-  }
-
-  if (req.cb) req.cb();
-
-  if (self._pendingWriteReqs == 0 && self._flags & FLAG_DESTROY_SOON) {
-    self._destroy();
-  }
+  if (req.cb)
+    req.cb.call(self);
 }
 
 
@@ -625,7 +744,7 @@ function connect(self, address, port, addressType, localAddress) {
     }
 
     if (r) {
-      self._destroy(errnoException(errno, 'bind'));
+      self._destroy(errnoException(process._errno, 'bind'));
       return;
     }
   }
@@ -642,12 +761,15 @@ function connect(self, address, port, addressType, localAddress) {
   if (connectReq !== null) {
     connectReq.oncomplete = afterConnect;
   } else {
-    self._destroy(errnoException(errno, 'connect'));
+    self._destroy(errnoException(process._errno, 'connect'));
   }
 }
 
 
 Socket.prototype.connect = function(options, cb) {
+  if (this.write !== Socket.prototype.write)
+    this.write = Socket.prototype.write;
+
   if (typeof options !== 'object') {
     // Old API:
     // connect(port, [host], [cb])
@@ -656,10 +778,20 @@ Socket.prototype.connect = function(options, cb) {
     return Socket.prototype.connect.apply(this, args);
   }
 
+  if (this.destroyed) {
+    this._readableState.reading = false;
+    this._readableState.ended = false;
+    this._writableState.ended = false;
+    this._writableState.ending = false;
+    this._writableState.finished = false;
+    this.destroyed = false;
+    this._handle = null;
+  }
+
   var self = this;
   var pipe = !!options.path;
 
-  if (this.destroyed || !this._handle) {
+  if (!this._handle) {
     this._handle = pipe ? createPipe() : createTCP();
     initSocketHandle(this);
   }
@@ -668,7 +800,7 @@ Socket.prototype.connect = function(options, cb) {
     self.once('connect', cb);
   }
 
-  timers.active(this);
+  timers._unrefActive(this);
 
   self._connecting = true;
   self.writable = true;
@@ -699,7 +831,7 @@ Socket.prototype.connect = function(options, cb) {
           self._destroy();
         });
       } else {
-        timers.active(self);
+        timers._unrefActive(self);
 
         addressType = addressType || 4;
 
@@ -715,6 +847,19 @@ Socket.prototype.connect = function(options, cb) {
 };
 
 
+Socket.prototype.ref = function() {
+  if (this._handle)
+    this._handle.ref();
+};
+
+
+Socket.prototype.unref = function() {
+  if (this._handle)
+    this._handle.unref();
+};
+
+
+
 function afterConnect(status, handle, req, readable, writable) {
   var self = handle.owner;
 
@@ -733,31 +878,18 @@ function afterConnect(status, handle, req, readable, writable) {
   if (status == 0) {
     self.readable = readable;
     self.writable = writable;
-    timers.active(self);
-
-    if (self.readable && !self._paused) {
-      handle.readStart();
-    }
-
-    if (self._connectQueue) {
-      debug('Drain the connect queue');
-      var connectQueue = self._connectQueue;
-      for (var i = 0; i < connectQueue.length; i++) {
-        self._write.apply(self, connectQueue[i]);
-      }
-      self._connectQueueCleanUp();
-    }
+    timers._unrefActive(self);
 
     self.emit('connect');
 
-    if (self._flags & FLAG_SHUTDOWN_QUEUED) {
-      // end called before connected - call end now with no data
-      self._flags &= ~FLAG_SHUTDOWN_QUEUED;
-      self.end();
-    }
+    // start the first read, or get an immediate EOF.
+    // this doesn't actually consume any bytes, because len=0.
+    if (readable)
+      self.read(0);
+
   } else {
-    self._connectQueueCleanUp();
-    self._destroy(errnoException(errno, 'connect'));
+    self._connecting = false;
+    self._destroy(errnoException(process._errno, 'connect'));
   }
 }
 
@@ -796,24 +928,25 @@ function Server(/* [ options, ] listener */) {
 
   this._connections = 0;
 
-  // when server is using slaves .connections is not reliable
-  // so null will be return if thats the case
   Object.defineProperty(this, 'connections', {
-    get: function() {
+    get: util.deprecate(function() {
+
       if (self._usingSlaves) {
         return null;
       }
       return self._connections;
-    },
-    set: function(val) {
+    }, 'connections property is deprecated. Use getConnections() method'),
+    set: util.deprecate(function(val) {
       return (self._connections = val);
-    },
+    }, 'connections property is deprecated. Use getConnections() method'),
     configurable: true, enumerable: true
   });
 
-  this.allowHalfOpen = options.allowHalfOpen || false;
-
   this._handle = null;
+  this._usingSlaves = false;
+  this._slaves = [];
+
+  this.allowHalfOpen = options.allowHalfOpen || false;
 }
 util.inherits(Server, events.EventEmitter);
 exports.Server = Server;
@@ -829,25 +962,18 @@ var createServerHandle = exports._createServerHandle =
   var handle;
 
   if (typeof fd === 'number' && fd >= 0) {
-    var tty_wrap = process.binding('tty_wrap');
-    var type = tty_wrap.guessHandleType(fd);
-    switch (type) {
-      case 'PIPE':
-        debug('listen pipe fd=' + fd);
-        // create a PipeWrap
-        handle = createPipe();
-        handle.open(fd);
-        handle.readable = true;
-        handle.writable = true;
-        break;
-
-      default:
-        // Not a fd we can listen on.  This will trigger an error.
-        debug('listen invalid fd=' + fd + ' type=' + type);
-        global.errno = 'EINVAL'; // hack, callers expect that errno is set
-        handle = null;
-        break;
+    try {
+      handle = createHandle(fd);
     }
+    catch (e) {
+      // Not a fd we can listen on.  This will trigger an error.
+      debug('listen invalid fd=' + fd + ': ' + e.message);
+      process._errno = 'EINVAL'; // hack, callers expect that errno is set
+      return null;
+    }
+    handle.open(fd);
+    handle.readable = true;
+    handle.writable = true;
     return handle;
 
   } else if (port == -1 && addressType == -1) {
@@ -881,20 +1007,24 @@ var createServerHandle = exports._createServerHandle =
 
 
 Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
+  debug('listen2', address, port, addressType, backlog);
   var self = this;
   var r = 0;
 
   // If there is not yet a handle, we need to create one and bind.
   // In the case of a server sent via IPC, we don't need to do this.
   if (!self._handle) {
+    debug('_listen2: create a handle');
     self._handle = createServerHandle(address, port, addressType, fd);
     if (!self._handle) {
-      var error = errnoException(errno, 'listen');
+      var error = errnoException(process._errno, 'listen');
       process.nextTick(function() {
         self.emit('error', error);
       });
       return;
     }
+  } else {
+    debug('_listen2: have a handle already');
   }
 
   self._handle.onconnection = onconnection;
@@ -906,7 +1036,7 @@ Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
   r = self._handle.listen(backlog || 511);
 
   if (r) {
-    var ex = errnoException(errno, 'listen');
+    var ex = errnoException(process._errno, 'listen');
     self._handle.close();
     self._handle = null;
     process.nextTick(function() {
@@ -927,14 +1057,34 @@ Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
 function listen(self, address, port, addressType, backlog, fd) {
   if (!cluster) cluster = require('cluster');
 
-  if (cluster.isWorker) {
-    cluster._getServer(self, address, port, addressType, fd, function(handle) {
-      self._handle = handle;
-      self._listen2(address, port, addressType, backlog, fd);
-    });
-  } else {
+  if (cluster.isMaster) {
     self._listen2(address, port, addressType, backlog, fd);
+    return;
   }
+
+  cluster._getServer(self, address, port, addressType, fd, function(handle,
+                                                                    err) {
+        // EACCESS and friends
+        if (err) {
+          self.emit('error', errnoException(err, 'bind'));
+          return;
+        }
+
+        // Some operating systems (notably OS X and Solaris) don't report
+        // EADDRINUSE errors right away. libuv mimics that behavior for the
+        // sake of platform consistency but that means we have have a socket on
+        // our hands that is not actually bound. That's why we check if the
+        // actual port matches what we requested and if not, raise an error.
+        // The exception is when port == 0 because that means "any random
+        // port".
+        if (port && handle.getsockname && port != handle.getsockname().port) {
+          self.emit('error', errnoException('EADDRINUSE', 'bind'));
+          return;
+        }
+
+        self._handle = handle;
+        self._listen2(address, port, addressType, backlog, fd);
+      });
 }
 
 
@@ -1014,7 +1164,7 @@ function onconnection(clientHandle) {
   debug('onconnection');
 
   if (!clientHandle) {
-    self.emit('error', errnoException(errno, 'accept'));
+    self.emit('error', errnoException(process._errno, 'accept'));
     return;
   }
 
@@ -1029,18 +1179,55 @@ function onconnection(clientHandle) {
   });
   socket.readable = socket.writable = true;
 
-  clientHandle.readStart();
 
   self._connections++;
   socket.server = self;
 
   DTRACE_NET_SERVER_CONNECTION(socket);
+  COUNTER_NET_SERVER_CONNECTION(socket);
   self.emit('connection', socket);
-  socket.emit('connect');
 }
 
 
+Server.prototype.getConnections = function(cb) {
+  function end(err, connections) {
+    process.nextTick(function() {
+      cb(err, connections);
+    });
+  }
+
+  if (!this._usingSlaves) {
+    return end(null, this._connections);
+  }
+
+  // Poll slaves
+  var left = this._slaves.length,
+      total = this._connections;
+
+  function oncount(err, count) {
+    if (err) {
+      left = -1;
+      return end(err);
+    }
+
+    total += count;
+    if (--left === 0) return end(null, total);
+  }
+
+  this._slaves.forEach(function(slave) {
+    slave.getConnections(oncount);
+  });
+};
+
+
 Server.prototype.close = function(cb) {
+  function onSlaveClose() {
+    if (--left !== 0) return;
+
+    self._connections = 0;
+    self._emitCloseIfDrained();
+  }
+
   if (!this._handle) {
     // Throw error. Follows net_legacy behaviour.
     throw new Error('Not running');
@@ -1051,25 +1238,38 @@ Server.prototype.close = function(cb) {
   }
   this._handle.close();
   this._handle = null;
-  this._emitCloseIfDrained();
 
-  // fetch new socket lists
   if (this._usingSlaves) {
-    this._slaves.forEach(function(socketList) {
-      if (socketList.list.length === 0) return;
-      socketList.update();
+    var self = this,
+        left = this._slaves.length;
+
+    // Increment connections to be sure that, even if all sockets will be closed
+    // during polling of slaves, `close` event will be emitted only once.
+    this._connections++;
+
+    // Poll slaves
+    this._slaves.forEach(function(slave) {
+      slave.close(onSlaveClose);
     });
+  } else {
+    this._emitCloseIfDrained();
   }
 
   return this;
 };
 
 Server.prototype._emitCloseIfDrained = function() {
+  debug('SERVER _emitCloseIfDrained');
   var self = this;
 
-  if (self._handle || self._connections) return;
+  if (self._handle || self._connections) {
+    debug('SERVER handle? %j   connections? %d',
+          !!self._handle, self._connections);
+    return;
+  }
 
   process.nextTick(function() {
+    debug('SERVER: emit close');
     self.emit('close');
   });
 };
@@ -1079,39 +1279,25 @@ Server.prototype.listenFD = util.deprecate(function(fd, type) {
   return this.listen({ fd: fd });
 }, 'listenFD is deprecated. Use listen({fd: <number>}).');
 
-// when sending a socket using fork IPC this function is executed
 Server.prototype._setupSlave = function(socketList) {
-  if (!this._usingSlaves) {
-    this._usingSlaves = true;
-    this._slaves = [];
-  }
+  this._usingSlaves = true;
   this._slaves.push(socketList);
 };
 
+Server.prototype.ref = function() {
+  if (this._handle)
+    this._handle.ref();
+};
+
+Server.prototype.unref = function() {
+  if (this._handle)
+    this._handle.unref();
+};
+
 
 // TODO: isIP should be moved to the DNS code. Putting it here now because
 // this is what the legacy system did.
-// NOTE: This does not accept IPv6 with an IPv4 dotted address at the end,
-//  and it does not detect more than one double : in a string.
-exports.isIP = function(input) {
-  if (!input) {
-    return 0;
-  } else if (/^(\d?\d?\d)\.(\d?\d?\d)\.(\d?\d?\d)\.(\d?\d?\d)$/.test(input)) {
-    var parts = input.split('.');
-    for (var i = 0; i < parts.length; i++) {
-      var part = parseInt(parts[i]);
-      if (part < 0 || 255 < part) {
-        return 0;
-      }
-    }
-    return 4;
-  } else if (/^::|^::1|^([a-fA-F0-9]{1,4}::?){1,7}([a-fA-F0-9]{1,4})$/.test(
-      input)) {
-    return 6;
-  } else {
-    return 0;
-  }
-};
+exports.isIP = cares.isIP;
 
 
 exports.isIPv4 = function(input) {
