@@ -22,7 +22,7 @@
 /* A repl library that you can include in your own code to get a runtime
  * interface to your program.
  *
- *   var repl = require("/repl.js");
+ *   var repl = require("repl");
  *   // start repl on stdin
  *   repl.start("prompt> ");
  *
@@ -47,6 +47,7 @@ var vm = require('vm');
 var path = require('path');
 var fs = require('fs');
 var rl = require('readline');
+var Console = require('console').Console;
 var EventEmitter = require('events').EventEmitter;
 
 // If obj.hasOwnProperty has been overridden, then calling
@@ -68,8 +69,8 @@ module.paths = require('module')._nodeModulePaths(module.filename);
 exports.writer = util.inspect;
 
 exports._builtinLibs = ['assert', 'buffer', 'child_process', 'cluster',
-  'crypto', 'dgram', 'dns', 'events', 'fs', 'http', 'https', 'net',
-  'os', 'path', 'punycode', 'querystring', 'readline', 'repl',
+  'crypto', 'dgram', 'dns', 'domain', 'events', 'fs', 'http', 'https', 'net',
+  'os', 'path', 'punycode', 'querystring', 'readline', 'stream',
   'string_decoder', 'tls', 'tty', 'url', 'util', 'vm', 'zlib'];
 
 
@@ -113,12 +114,15 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     } catch (e) {
       err = e;
     }
-    cb(err, result);
+    if (err && process.domain) {
+      process.domain.emit('error', err);
+      process.domain.exit();
+    }
+    else {
+      cb(err, result);
+    }
   };
 
-  self.resetContext();
-  self.bufferedCommand = '';
-
   if (!input && !output) {
     // legacy API, passing a 'stream'/'socket' option
     if (!stream) {
@@ -136,10 +140,12 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     }
   }
 
-
   self.inputStream = input;
   self.outputStream = output;
 
+  self.resetContext();
+  self.bufferedCommand = '';
+
   self.prompt = (prompt != undefined ? prompt : '> ');
 
   function complete(text, callback) {
@@ -206,7 +212,7 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
 
     // Check to see if a REPL keyword was used. If it returns true,
     // display next prompt and return.
-    if (cmd && cmd.charAt(0) === '.') {
+    if (cmd && cmd.charAt(0) === '.' && isNaN(parseFloat(cmd))) {
       var matches = cmd.match(/^(\.[^\s]+)\s*(.*)$/);
       var keyword = matches && matches[1];
       var rest = matches && matches[2];
@@ -218,21 +224,6 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
       }
     }
 
-    // Check if a builtin module name was used and then include it
-    // if there's no conflict.
-    if (exports._builtinLibs.indexOf(cmd) !== -1) {
-      var lib = require(cmd);
-      if (cmd in self.context && lib !== self.context[cmd]) {
-        self.outputStream.write('A different "' + cmd +
-                                '" already exists globally\n');
-      } else {
-        self.context._ = self.context[cmd] = lib;
-        self.outputStream.write(self.writer(lib) + '\n');
-      }
-      self.displayPrompt();
-      return;
-    }
-
     if (!skipCatchall) {
       var evalCmd = self.bufferedCommand + cmd + '\n';
 
@@ -275,7 +266,7 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
           self.outputStream.write('npm should be run outside of the ' +
                                   'node repl, in your normal shell.\n' +
                                   '(Press Control-D to exit.)\n');
-          self.bufferedCmd = '';
+          self.bufferedCommand = '';
           self.displayPrompt();
           return;
         }
@@ -331,21 +322,39 @@ REPLServer.prototype.createContext = function() {
   } else {
     context = vm.createContext();
     for (var i in global) context[i] = global[i];
+    context.console = new Console(this.outputStream);
+    context.global = context;
+    context.global.global = context;
   }
 
   context.module = module;
   context.require = require;
-  context.global = context;
-  context.global.global = context;
 
   this.lines = [];
   this.lines.level = [];
 
+  // make built-in modules available directly
+  // (loaded lazily)
+  exports._builtinLibs.forEach(function(name) {
+    Object.defineProperty(context, name, {
+      get: function() {
+        var lib = require(name);
+        context._ = context[name] = lib;
+        return lib;
+      },
+      // allow the creation of other globals with this name
+      set: function(val) {
+        delete context[name];
+        context[name] = val;
+      },
+      configurable: true
+    });
+  });
+
   return context;
 };
 
 REPLServer.prototype.resetContext = function() {
-  for (var i in require.cache) delete require.cache[i];
   this.context = this.createContext();
 };
 
@@ -437,11 +446,7 @@ REPLServer.prototype.complete = function(line, callback) {
     completionGroupsLoaded();
   } else if (match = line.match(requireRE)) {
     // require('...<Tab>')
-    //TODO: suggest require.exts be exposed to be introspec registered
-    //extensions?
-    //TODO: suggest include the '.' in exts in internal repr: parity with
-    //`path.extname`.
-    var exts = ['.js', '.node'];
+    var exts = Object.keys(require.extensions);
     var indexRe = new RegExp('^index(' + exts.map(regexpEscape).join('|') +
                              ')$');
 
@@ -636,7 +641,7 @@ REPLServer.prototype.complete = function(line, callback) {
         group.sort();
         for (var j = 0; j < group.length; j++) {
           c = group[j];
-          if (!hasOwnProperty(c)) {
+          if (!hasOwnProperty(uniq, c)) {
             completions.push(c);
             uniq[c] = true;
           }
@@ -915,7 +920,8 @@ function isSyntaxError(e) {
       // RegExp syntax error
       !e.match(/^SyntaxError: Invalid regular expression/) &&
       !e.match(/^SyntaxError: Invalid flags supplied to RegExp constructor/) &&
+      // "strict mode" syntax errors
+      !e.match(/^SyntaxError: .*strict mode.*/i) &&
       // JSON.parse() error
-      !(e.match(/^SyntaxError: Unexpected (token .*|end of input)/) &&
-      e.match(/\n    at Object.parse \(native\)\n/));
+      !e.match(/\n {4}at Object.parse \(native\)\n/);
 }
