@@ -19,33 +19,57 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+function assertEncoding(encoding) {
+  if (encoding && !Buffer.isEncoding(encoding)) {
+    throw new Error('Unknown encoding: ' + encoding);
+  }
+}
+
 var StringDecoder = exports.StringDecoder = function(encoding) {
   this.encoding = (encoding || 'utf8').toLowerCase().replace(/[-_]/, '');
-  if (this.encoding === 'utf8') {
-    this.charBuffer = new Buffer(4);
-    this.charReceived = 0;
-    this.charLength = 0;
+  assertEncoding(encoding);
+  switch (this.encoding) {
+    case 'utf8':
+      // CESU-8 represents each of Surrogate Pair by 3-bytes
+      this.surrogateSize = 3;
+      break;
+    case 'ucs2':
+    case 'utf16le':
+      // UTF-16 represents each of Surrogate Pair by 2-bytes
+      this.surrogateSize = 2;
+      this.detectIncompleteChar = utf16DetectIncompleteChar;
+      break;
+    case 'base64':
+      // Base-64 stores 3 bytes in 4 chars, and pads the remainder.
+      this.surrogateSize = 3;
+      this.detectIncompleteChar = base64DetectIncompleteChar;
+      break;
+    default:
+      this.write = passThroughWrite;
+      return;
   }
+
+  this.charBuffer = new Buffer(6);
+  this.charReceived = 0;
+  this.charLength = 0;
 };
 
 
 StringDecoder.prototype.write = function(buffer) {
-  // If not utf8...
-  if (this.encoding !== 'utf8') {
-    return buffer.toString(this.encoding);
-  }
-
   var charStr = '';
+  var offset = 0;
+
   // if our last write ended with an incomplete multibyte character
-  if (this.charLength) {
+  while (this.charLength) {
     // determine how many remaining bytes this buffer has to offer for this char
     var i = (buffer.length >= this.charLength - this.charReceived) ?
                 this.charLength - this.charReceived :
                 buffer.length;
 
     // add the new bytes to the char buffer
-    buffer.copy(this.charBuffer, this.charReceived, 0, i);
-    this.charReceived += i;
+    buffer.copy(this.charBuffer, this.charReceived, offset, i);
+    this.charReceived += (i - offset);
+    offset = i;
 
     if (this.charReceived < this.charLength) {
       // still not enough chars in this buffer? wait for more ...
@@ -53,7 +77,15 @@ StringDecoder.prototype.write = function(buffer) {
     }
 
     // get the character that was split
-    charStr = this.charBuffer.slice(0, this.charLength).toString();
+    charStr = this.charBuffer.slice(0, this.charLength).toString(this.encoding);
+
+    // lead surrogate (D800-DBFF) is also the incomplete character
+    var charCode = charStr.charCodeAt(charStr.length - 1);
+    if (charCode >= 0xD800 && charCode <= 0xDBFF) {
+      this.charLength += this.surrogateSize;
+      charStr = '';
+      continue;
+    }
     this.charReceived = this.charLength = 0;
 
     // if there are no more bytes in this buffer, just emit our char
@@ -61,9 +93,38 @@ StringDecoder.prototype.write = function(buffer) {
 
     // otherwise cut off the characters end from the beginning of this buffer
     buffer = buffer.slice(i, buffer.length);
+    break;
   }
 
+  var lenIncomplete = this.detectIncompleteChar(buffer);
 
+  var end = buffer.length;
+  if (this.charLength) {
+    // buffer the incomplete character bytes we got
+    buffer.copy(this.charBuffer, 0, buffer.length - lenIncomplete, end);
+    this.charReceived = lenIncomplete;
+    end -= lenIncomplete;
+  }
+
+  charStr += buffer.toString(this.encoding, 0, end);
+
+  var end = charStr.length - 1;
+  var charCode = charStr.charCodeAt(end);
+  // lead surrogate (D800-DBFF) is also the incomplete character
+  if (charCode >= 0xD800 && charCode <= 0xDBFF) {
+    var size = this.surrogateSize;
+    this.charLength += size;
+    this.charReceived += size;
+    this.charBuffer.copy(this.charBuffer, size, 0, size);
+    this.charBuffer.write(charStr.charAt(charStr.length - 1), this.encoding);
+    return charStr.substring(0, end);
+  }
+
+  // or just emit the charStr
+  return charStr;
+};
+
+StringDecoder.prototype.detectIncompleteChar = function(buffer) {
   // determine how many bytes we have to check at the end of this buffer
   var i = (buffer.length >= 3) ? 3 : buffer.length;
 
@@ -93,20 +154,36 @@ StringDecoder.prototype.write = function(buffer) {
     }
   }
 
-  if (!this.charLength) {
-    // no incomplete char at the end of this buffer, emit the whole thing
-    return charStr + buffer.toString();
-  }
-
-  // buffer the incomplete character bytes we got
-  buffer.copy(this.charBuffer, 0, buffer.length - i, buffer.length);
-  this.charReceived = i;
-
-  if (buffer.length - i > 0) {
-    // buffer had more bytes before the incomplete char, emit them
-    return charStr + buffer.toString('utf8', 0, buffer.length - i);
-  }
-
-  // or just emit the charStr
-  return charStr;
+  return i;
 };
+
+StringDecoder.prototype.end = function(buffer) {
+  var res = '';
+  if (buffer && buffer.length)
+    res = this.write(buffer);
+
+  if (this.charReceived) {
+    var cr = this.charReceived;
+    var buf = this.charBuffer;
+    var enc = this.encoding;
+    res += buf.slice(0, cr).toString(enc);
+  }
+
+  return res;
+};
+
+function passThroughWrite(buffer) {
+  return buffer.toString(this.encoding);
+}
+
+function utf16DetectIncompleteChar(buffer) {
+  var incomplete = this.charReceived = buffer.length % 2;
+  this.charLength = incomplete ? 2 : 0;
+  return incomplete;
+}
+
+function base64DetectIncompleteChar(buffer) {
+  var incomplete = this.charReceived = buffer.length % 3;
+  this.charLength = incomplete ? 3 : 0;
+  return incomplete;
+}
