@@ -33,12 +33,12 @@ exports.format = function(f) {
   var args = arguments;
   var len = args.length;
   var str = String(f).replace(formatRegExp, function(x) {
+    if (x === '%%') return '%';
     if (i >= len) return x;
     switch (x) {
       case '%s': return String(args[i++]);
       case '%d': return Number(args[i++]);
       case '%j': return JSON.stringify(args[i++]);
-      case '%%': return '%';
       default:
         return x;
     }
@@ -54,6 +54,33 @@ exports.format = function(f) {
 };
 
 
+// Mark that a method should not be used.
+// Returns a modified function which warns once by default.
+// If --no-deprecation is set, then it is a no-op.
+exports.deprecate = function(fn, msg) {
+  if (process.noDeprecation === true) {
+    return fn;
+  }
+
+  var warned = false;
+  function deprecated() {
+    if (!warned) {
+      if (process.throwDeprecation) {
+        throw new Error(msg);
+      } else if (process.traceDeprecation) {
+        console.trace(msg);
+      } else {
+        console.error(msg);
+      }
+      warned = true;
+    }
+    return fn.apply(this, arguments);
+  }
+
+  return deprecated;
+};
+
+
 exports.print = function() {
   for (var i = 0, len = arguments.length; i < len; ++i) {
     process.stdout.write(String(arguments[i]));
@@ -85,25 +112,38 @@ var error = exports.error = function(x) {
  * in the best way possible given the different types.
  *
  * @param {Object} obj The object to print out.
- * @param {Boolean} showHidden Flag that shows hidden (not enumerable)
- *    properties of objects.
- * @param {Number} depth Depth in which to descend in object. Default is 2.
- * @param {Boolean} colors Flag to turn on ANSI escape codes to color the
- *    output. Default is false (no coloring).
+ * @param {Object} opts Optional options object that alters the output.
  */
-function inspect(obj, showHidden, depth, colors) {
+/* legacy: obj, showHidden, depth, colors*/
+function inspect(obj, opts) {
+  // default options
   var ctx = {
-    showHidden: showHidden,
     seen: [],
-    stylize: colors ? stylizeWithColor : stylizeNoColor
+    stylize: stylizeNoColor
   };
-  return formatValue(ctx, obj, (typeof depth === 'undefined' ? 2 : depth));
+  // legacy...
+  if (arguments.length >= 3) ctx.depth = arguments[2];
+  if (arguments.length >= 4) ctx.colors = arguments[3];
+  if (typeof opts === 'boolean') {
+    // legacy...
+    ctx.showHidden = opts;
+  } else if (opts) {
+    // got an "options" object
+    exports._extend(ctx, opts);
+  }
+  // set default options
+  if (typeof ctx.showHidden === 'undefined') ctx.showHidden = false;
+  if (typeof ctx.depth === 'undefined') ctx.depth = 2;
+  if (typeof ctx.colors === 'undefined') ctx.colors = false;
+  if (typeof ctx.customInspect === 'undefined') ctx.customInspect = true;
+  if (ctx.colors) ctx.stylize = stylizeWithColor;
+  return formatValue(ctx, obj, ctx.depth);
 }
 exports.inspect = inspect;
 
 
 // http://en.wikipedia.org/wiki/ANSI_escape_code#graphics
-var colors = {
+inspect.colors = {
   'bold' : [1, 22],
   'italic' : [3, 23],
   'underline' : [4, 24],
@@ -120,7 +160,7 @@ var colors = {
 };
 
 // Don't use 'blue' not visible on cmd.exe
-var styles = {
+inspect.styles = {
   'special': 'cyan',
   'number': 'yellow',
   'boolean': 'yellow',
@@ -134,11 +174,11 @@ var styles = {
 
 
 function stylizeWithColor(str, styleType) {
-  var style = styles[styleType];
+  var style = inspect.styles[styleType];
 
   if (style) {
-    return '\033[' + colors[style][0] + 'm' + str +
-           '\033[' + colors[style][1] + 'm';
+    return '\u001b[' + inspect.colors[style][0] + 'm' + str +
+           '\u001b[' + inspect.colors[style][1] + 'm';
   } else {
     return str;
   }
@@ -150,15 +190,26 @@ function stylizeNoColor(str, styleType) {
 }
 
 
+function arrayToHash(array) {
+  var hash = {};
+
+  array.forEach(function(val, idx) {
+    hash[val] = true;
+  });
+
+  return hash;
+}
+
+
 function formatValue(ctx, value, recurseTimes) {
   // Provide a hook for user-specified inspect functions.
   // Check that value is an object with an inspect function on it
-  if (value && typeof value.inspect === 'function' &&
+  if (ctx.customInspect && value && typeof value.inspect === 'function' &&
       // Filter out the util module, it's inspect function is special
       value.inspect !== exports.inspect &&
       // Also filter out any prototype objects using the circular check.
       !(value.constructor && value.constructor.prototype === value)) {
-    return value.inspect(recurseTimes);
+    return String(value.inspect(recurseTimes));
   }
 
   // Primitive types cannot have properties
@@ -168,8 +219,12 @@ function formatValue(ctx, value, recurseTimes) {
   }
 
   // Look up the keys of the object.
-  var visibleKeys = Object.keys(value);
-  var keys = ctx.showHidden ? Object.getOwnPropertyNames(value) : visibleKeys;
+  var keys = Object.keys(value);
+  var visibleKeys = arrayToHash(keys);
+
+  if (ctx.showHidden) {
+    keys = Object.getOwnPropertyNames(value);
+  }
 
   // Some type of object without properties can be shortcutted.
   if (keys.length === 0) {
@@ -181,7 +236,7 @@ function formatValue(ctx, value, recurseTimes) {
       return ctx.stylize(RegExp.prototype.toString.call(value), 'regexp');
     }
     if (isDate(value)) {
-      return ctx.stylize(Date.prototype.toUTCString.call(value), 'date');
+      return ctx.stylize(Date.prototype.toString.call(value), 'date');
     }
     if (isError(value)) {
       return formatError(value);
@@ -278,7 +333,7 @@ function formatError(value) {
 function formatArray(ctx, value, recurseTimes, visibleKeys, keys) {
   var output = [];
   for (var i = 0, l = value.length; i < l; ++i) {
-    if (Object.prototype.hasOwnProperty.call(value, String(i))) {
+    if (hasOwnProperty(value, String(i))) {
       output.push(formatProperty(ctx, value, recurseTimes, visibleKeys,
           String(i), true));
     } else {
@@ -309,7 +364,7 @@ function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
       str = ctx.stylize('[Setter]', 'special');
     }
   }
-  if (visibleKeys.indexOf(key) < 0) {
+  if (!hasOwnProperty(visibleKeys, key)) {
     name = '[' + key + ']';
   }
   if (!str) {
@@ -407,18 +462,11 @@ function objectToString(o) {
 }
 
 
-var pWarning;
-
-exports.p = function() {
-  if (!pWarning) {
-    pWarning = 'util.p will be removed in future versions of Node. ' +
-               'Use util.puts(util.inspect()) instead.\n';
-    exports.error(pWarning);
-  }
+exports.p = exports.deprecate(function() {
   for (var i = 0, len = arguments.length; i < len; ++i) {
     error(exports.inspect(arguments[i]));
   }
-};
+}, 'util.p: Use console.error() instead.');
 
 
 function pad(n) {
@@ -444,18 +492,12 @@ exports.log = function(msg) {
 };
 
 
-var execWarning;
-exports.exec = function() {
-  if (!execWarning) {
-    execWarning = 'util.exec has moved to the "child_process" module.' +
-                  ' Please update your source code.';
-    error(execWarning);
-  }
+exports.exec = exports.deprecate(function() {
   return require('child_process').exec.apply(this, arguments);
-};
+}, 'util.exec is now called `child_process.exec`.');
 
 
-exports.pump = function(readStream, writeStream, callback) {
+function pump(readStream, writeStream, callback) {
   var callbackCalled = false;
 
   function call(a, b, c) {
@@ -490,7 +532,9 @@ exports.pump = function(readStream, writeStream, callback) {
     readStream.destroy();
     call(err);
   });
-};
+}
+exports.pump = exports.deprecate(pump,
+    'util.pump() is deprecated. Use readableStream.pipe() instead.');
 
 
 /**
@@ -498,7 +542,7 @@ exports.pump = function(readStream, writeStream, callback) {
  *
  * The Function.prototype.inherits from lang.js rewritten as a standalone
  * function (not on Function.prototype). NOTE: If this file is to be loaded
- * during bootstrapping this function needs to be revritten using some native
+ * during bootstrapping this function needs to be rewritten using some native
  * functions as prototype setup using normal JavaScript does not work as
  * expected during bootstrapping (see mirror.js in r114903).
  *
@@ -518,18 +562,18 @@ exports.inherits = function(ctor, superCtor) {
   });
 };
 
-var deprecationWarnings;
+exports._extend = function(origin, add) {
+  // Don't do anything if add isn't an object
+  if (!add || typeof add !== 'object') return origin;
 
-exports._deprecationWarning = function(moduleId, message) {
-  if (!deprecationWarnings)
-    deprecationWarnings = {};
-  else if (message in deprecationWarnings)
-    return;
-
-  deprecationWarnings[message] = true;
-
-  if ((new RegExp('\\b' + moduleId + '\\b')).test(process.env.NODE_DEBUG))
-    console.trace(message);
-  else
-    console.error(message);
+  var keys = Object.keys(add);
+  var i = keys.length;
+  while (i--) {
+    origin[keys[i]] = add[keys[i]];
+  }
+  return origin;
 };
+
+function hasOwnProperty(obj, prop) {
+  return Object.prototype.hasOwnProperty.call(obj, prop);
+}
