@@ -20,33 +20,23 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 #include <assert.h>
-#include <node.h>
-#include <req_wrap.h>
-#include <uv.h>
-
+#include <errno.h>
+#include <stdlib.h>
 #include <string.h>
 
+#define CARES_STATICLIB
+#include "ares.h"
+#include "node.h"
+#include "req_wrap.h"
+#include "tree.h"
+#include "uv.h"
+
 #if defined(__OpenBSD__) || defined(__MINGW32__) || defined(_MSC_VER)
 # include <nameser.h>
 #else
 # include <arpa/nameser.h>
 #endif
 
-// Temporary hack: libuv should provide uv_inet_pton and uv_inet_ntop.
-#if defined(__MINGW32__) || defined(_MSC_VER)
-  extern "C" {
-#   include <inet_net_pton.h>
-#   include <inet_ntop.h>
-  }
-# define uv_inet_pton ares_inet_pton
-# define uv_inet_ntop ares_inet_ntop
-
-#else // __POSIX__
-# include <arpa/inet.h>
-# define uv_inet_pton inet_pton
-# define uv_inet_ntop inet_ntop
-#endif
-
 
 namespace node {
 
@@ -69,9 +59,139 @@ using v8::Value;
 
 typedef class ReqWrap<uv_getaddrinfo_t> GetAddrInfoReqWrap;
 
-static Persistent<String> oncomplete_sym;
+struct ares_task_t {
+  UV_HANDLE_FIELDS
+  ares_socket_t sock;
+  uv_poll_t poll_watcher;
+  RB_ENTRY(ares_task_t) node;
+};
 
+
+static Persistent<String> oncomplete_sym;
 static ares_channel ares_channel;
+static uv_timer_t ares_timer;
+static RB_HEAD(ares_task_list, ares_task_t) ares_tasks;
+
+
+static int cmp_ares_tasks(const ares_task_t* a, const ares_task_t* b) {
+  if (a->sock < b->sock) return -1;
+  if (a->sock > b->sock) return 1;
+  return 0;
+}
+
+
+RB_GENERATE_STATIC(ares_task_list, ares_task_t, node, cmp_ares_tasks)
+
+
+
+/* This is called once per second by loop->timer. It is used to constantly */
+/* call back into c-ares for possibly processing timeouts. */
+static void ares_timeout(uv_timer_t* handle, int status) {
+  assert(!RB_EMPTY(&ares_tasks));
+  ares_process_fd(ares_channel, ARES_SOCKET_BAD, ARES_SOCKET_BAD);
+}
+
+
+static void ares_poll_cb(uv_poll_t* watcher, int status, int events) {
+  ares_task_t* task = container_of(watcher, ares_task_t, poll_watcher);
+
+  /* Reset the idle timer */
+  uv_timer_again(&ares_timer);
+
+  if (status < 0) {
+    /* An error happened. Just pretend that the socket is both readable and */
+    /* writable. */
+    ares_process_fd(ares_channel, task->sock, task->sock);
+    return;
+  }
+
+  /* Process DNS responses */
+  ares_process_fd(ares_channel,
+                  events & UV_READABLE ? task->sock : ARES_SOCKET_BAD,
+                  events & UV_WRITABLE ? task->sock : ARES_SOCKET_BAD);
+}
+
+
+static void ares_poll_close_cb(uv_handle_t* watcher) {
+  ares_task_t* task = container_of(watcher, ares_task_t, poll_watcher);
+  free(task);
+}
+
+
+/* Allocates and returns a new ares_task_t */
+static ares_task_t* ares_task_create(uv_loop_t* loop, ares_socket_t sock) {
+  ares_task_t* task = (ares_task_t*) malloc(sizeof *task);
+
+  if (task == NULL) {
+    /* Out of memory. */
+    return NULL;
+  }
+
+  task->loop = loop;
+  task->sock = sock;
+
+  if (uv_poll_init_socket(loop, &task->poll_watcher, sock) < 0) {
+    /* This should never happen. */
+    free(task);
+    return NULL;
+  }
+
+  return task;
+}
+
+
+/* Callback from ares when socket operation is started */
+static void ares_sockstate_cb(void* data, ares_socket_t sock,
+    int read, int write) {
+  uv_loop_t* loop = (uv_loop_t*) data;
+  ares_task_t* task;
+
+  ares_task_t lookup_task;
+  lookup_task.sock = sock;
+  task = RB_FIND(ares_task_list, &ares_tasks, &lookup_task);
+
+  if (read || write) {
+    if (!task) {
+      /* New socket */
+
+      /* If this is the first socket then start the timer. */
+      if (!uv_is_active((uv_handle_t*) &ares_timer)) {
+        assert(RB_EMPTY(&ares_tasks));
+        uv_timer_start(&ares_timer, ares_timeout, 1000, 1000);
+      }
+
+      task = ares_task_create(loop, sock);
+      if (task == NULL) {
+        /* This should never happen unless we're out of memory or something */
+        /* is seriously wrong. The socket won't be polled, but the the query */
+        /* will eventually time out. */
+        return;
+      }
+
+      RB_INSERT(ares_task_list, &ares_tasks, task);
+    }
+
+    /* This should never fail. If it fails anyway, the query will eventually */
+    /* time out. */
+    uv_poll_start(&task->poll_watcher,
+                  (read ? UV_READABLE : 0) | (write ? UV_WRITABLE : 0),
+                  ares_poll_cb);
+
+  } else {
+    /* read == 0 and write == 0 this is c-ares's way of notifying us that */
+    /* the socket is now closed. We must free the data associated with */
+    /* socket. */
+    assert(task &&
+           "When an ares socket is closed we should have a handle for it");
+
+    RB_REMOVE(ares_task_list, &ares_tasks, task);
+    uv_close((uv_handle_t*) &task->poll_watcher, ares_poll_close_cb);
+
+    if (RB_EMPTY(&ares_tasks)) {
+      uv_timer_stop(&ares_timer);
+    }
+  }
+}
 
 
 static Local<Array> HostentToAddresses(struct hostent* host) {
@@ -141,9 +261,9 @@ static const char* AresErrnoString(int errorno) {
 
 static void SetAresErrno(int errorno) {
   HandleScope scope;
-  Handle<Value> key = String::NewSymbol("errno");
-  Handle<Value> value = String::NewSymbol(AresErrnoString(errorno));
-  Context::GetCurrent()->Global()->Set(key, value);
+  Local<Value> key = String::NewSymbol("_errno");
+  Local<Value> value = String::NewSymbol(AresErrnoString(errorno));
+  node::process->Set(key, value);
 }
 
 
@@ -191,7 +311,7 @@ class QueryWrap {
 
   static void Callback(void *arg, int status, int timeouts,
       unsigned char* answer_buf, int answer_len) {
-    QueryWrap* wrap = reinterpret_cast<QueryWrap*>(arg);
+    QueryWrap* wrap = static_cast<QueryWrap*>(arg);
 
     if (status != ARES_SUCCESS) {
       wrap->ParseError(status);
@@ -204,7 +324,7 @@ class QueryWrap {
 
   static void Callback(void *arg, int status, int timeouts,
       struct hostent* host) {
-    QueryWrap* wrap = reinterpret_cast<QueryWrap*>(arg);
+    QueryWrap* wrap = static_cast<QueryWrap*>(arg);
 
     if (status != ARES_SUCCESS) {
       wrap->ParseError(status);
@@ -218,13 +338,13 @@ class QueryWrap {
   void CallOnComplete(Local<Value> answer) {
     HandleScope scope;
     Local<Value> argv[2] = { Integer::New(0), answer };
-    MakeCallback(object_, "oncomplete", 2, argv);
+    MakeCallback(object_, oncomplete_sym, ARRAY_SIZE(argv), argv);
   }
 
   void CallOnComplete(Local<Value> answer, Local<Value> family) {
     HandleScope scope;
     Local<Value> argv[3] = { Integer::New(0), answer, family };
-    MakeCallback(object_, "oncomplete", 3, argv);
+    MakeCallback(object_, oncomplete_sym, ARRAY_SIZE(argv), argv);
   }
 
   void ParseError(int status) {
@@ -233,7 +353,7 @@ class QueryWrap {
 
     HandleScope scope;
     Local<Value> argv[1] = { Integer::New(-1) };
-    MakeCallback(object_, "oncomplete", 1, argv);
+    MakeCallback(object_, oncomplete_sym, ARRAY_SIZE(argv), argv);
   }
 
   // Subclasses should implement the appropriate Parse method.
@@ -371,7 +491,8 @@ class QueryMxWrap: public QueryWrap {
          mx_current = mx_current->next) {
       Local<Object> mx_record = Object::New();
       mx_record->Set(exchange_symbol, String::New(mx_current->host));
-      mx_record->Set(priority_symbol, Integer::New(mx_current->priority));
+      mx_record->Set(priority_symbol,
+                     Integer::New(mx_current->priority));
       mx_records->Set(Integer::New(i++), mx_record);
     }
 
@@ -473,9 +594,12 @@ class QuerySrvWrap: public QueryWrap {
          srv_current = srv_current->next) {
       Local<Object> srv_record = Object::New();
       srv_record->Set(name_symbol, String::New(srv_current->host));
-      srv_record->Set(port_symbol, Integer::New(srv_current->port));
-      srv_record->Set(priority_symbol, Integer::New(srv_current->priority));
-      srv_record->Set(weight_symbol, Integer::New(srv_current->weight));
+      srv_record->Set(port_symbol,
+                      Integer::New(srv_current->port));
+      srv_record->Set(priority_symbol,
+                      Integer::New(srv_current->priority));
+      srv_record->Set(weight_symbol,
+                      Integer::New(srv_current->weight));
       srv_records->Set(Integer::New(i++), srv_record);
     }
 
@@ -485,6 +609,66 @@ class QuerySrvWrap: public QueryWrap {
   }
 };
 
+class QueryNaptrWrap: public QueryWrap {
+ public:
+  int Send(const char* name) {
+    ares_query(ares_channel,
+               name,
+               ns_c_in,
+               ns_t_naptr,
+               Callback,
+               GetQueryArg());
+    return 0;
+  }
+
+ protected:
+  void Parse(unsigned char* buf, int len) {
+    HandleScope scope;
+
+    ares_naptr_reply* naptr_start;
+    int status = ares_parse_naptr_reply(buf, len, &naptr_start);
+
+    if (status != ARES_SUCCESS) {
+      this->ParseError(status);
+      return;
+    }
+
+    Local<Array> naptr_records = Array::New();
+    Local<String> flags_symbol = String::NewSymbol("flags");
+    Local<String> service_symbol = String::NewSymbol("service");
+    Local<String> regexp_symbol = String::NewSymbol("regexp");
+    Local<String> replacement_symbol = String::NewSymbol("replacement");
+    Local<String> order_symbol = String::NewSymbol("order");
+    Local<String> preference_symbol = String::NewSymbol("preference");
+
+    int i = 0;
+    for (ares_naptr_reply* naptr_current = naptr_start;
+         naptr_current;
+         naptr_current = naptr_current->next) {
+
+      Local<Object> naptr_record = Object::New();
+
+      naptr_record->Set(flags_symbol,
+          String::New(reinterpret_cast<char*>(naptr_current->flags)));
+      naptr_record->Set(service_symbol,
+          String::New(reinterpret_cast<char*>(naptr_current->service)));
+      naptr_record->Set(regexp_symbol,
+          String::New(reinterpret_cast<char*>(naptr_current->regexp)));
+      naptr_record->Set(replacement_symbol,
+          String::New(naptr_current->replacement));
+      naptr_record->Set(order_symbol, Integer::New(naptr_current->order));
+      naptr_record->Set(preference_symbol,
+          Integer::New(naptr_current->preference));
+
+      naptr_records->Set(Integer::New(i++), naptr_record);
+    }
+
+    ares_free_data(naptr_start);
+
+    this->CallOnComplete(naptr_records);
+  }
+};
+
 
 class GetHostByAddrWrap: public QueryWrap {
  public:
@@ -492,10 +676,10 @@ class GetHostByAddrWrap: public QueryWrap {
     int length, family;
     char address_buffer[sizeof(struct in6_addr)];
 
-    if (uv_inet_pton(AF_INET, name, &address_buffer) == 1) {
+    if (uv_inet_pton(AF_INET, name, &address_buffer).code == UV_OK) {
       length = sizeof(struct in_addr);
       family = AF_INET;
-    } else if (uv_inet_pton(AF_INET6, name, &address_buffer) == 1) {
+    } else if (uv_inet_pton(AF_INET6, name, &address_buffer).code == UV_OK) {
       length = sizeof(struct in6_addr);
       family = AF_INET6;
     } else {
@@ -555,7 +739,7 @@ static Handle<Value> Query(const Arguments& args) {
   // object reference, causing wrap->GetObject() to return undefined.
   Local<Object> object = Local<Object>::New(wrap->GetObject());
 
-  String::Utf8Value name(args[0]->ToString());
+  String::Utf8Value name(args[0]);
 
   int r = wrap->Send(*name);
   if (r) {
@@ -584,7 +768,7 @@ static Handle<Value> QueryWithFamily(const Arguments& args) {
   // object reference, causing wrap->GetObject() to return undefined.
   Local<Object> object = Local<Object>::New(wrap->GetObject());
 
-  String::Utf8Value name(args[0]->ToString());
+  String::Utf8Value name(args[0]);
   int family = args[1]->Int32Value();
 
   int r = wrap->Send(*name, family);
@@ -637,11 +821,15 @@ void AfterGetAddrInfo(uv_getaddrinfo_t* req, int status, struct addrinfo* res) {
       if (address->ai_family == AF_INET) {
         // Juggle pointers
         addr = (char*) &((struct sockaddr_in*) address->ai_addr)->sin_addr;
-        const char* c = uv_inet_ntop(address->ai_family, addr, ip,
-            INET6_ADDRSTRLEN);
+        uv_err_t err = uv_inet_ntop(address->ai_family,
+                                    addr,
+                                    ip,
+                                    INET6_ADDRSTRLEN);
+        if (err.code != UV_OK)
+          continue;
 
         // Create JavaScript string
-        Local<String> s = String::New(c);
+        Local<String> s = String::New(ip);
         results->Set(n, s);
         n++;
       }
@@ -659,11 +847,15 @@ void AfterGetAddrInfo(uv_getaddrinfo_t* req, int status, struct addrinfo* res) {
       if (address->ai_family == AF_INET6) {
         // Juggle pointers
         addr = (char*) &((struct sockaddr_in6*) address->ai_addr)->sin6_addr;
-        const char* c = uv_inet_ntop(address->ai_family, addr, ip,
-            INET6_ADDRSTRLEN);
+        uv_err_t err = uv_inet_ntop(address->ai_family,
+                                    addr,
+                                    ip,
+                                    INET6_ADDRSTRLEN);
+        if (err.code != UV_OK)
+          continue;
 
         // Create JavaScript string
-        Local<String> s = String::New(c);
+        Local<String> s = String::New(ip);
         results->Set(n, s);
         n++;
       }
@@ -679,16 +871,34 @@ void AfterGetAddrInfo(uv_getaddrinfo_t* req, int status, struct addrinfo* res) {
   uv_freeaddrinfo(res);
 
   // Make the callback into JavaScript
-  MakeCallback(req_wrap->object_, "oncomplete", 1, argv);
+  MakeCallback(req_wrap->object_, oncomplete_sym, ARRAY_SIZE(argv), argv);
 
   delete req_wrap;
 }
 
 
+static Handle<Value> IsIP(const Arguments& args) {
+  HandleScope scope;
+
+  String::AsciiValue ip(args[0]);
+  char address_buffer[sizeof(struct in6_addr)];
+
+  if (uv_inet_pton(AF_INET, *ip, &address_buffer).code == UV_OK) {
+    return scope.Close(v8::Integer::New(4));
+  }
+
+  if (uv_inet_pton(AF_INET6, *ip, &address_buffer).code == UV_OK) {
+    return scope.Close(v8::Integer::New(6));
+  }
+
+  return scope.Close(v8::Integer::New(0));
+}
+
+
 static Handle<Value> GetAddrInfo(const Arguments& args) {
   HandleScope scope;
 
-  String::Utf8Value hostname(args[0]->ToString());
+  String::Utf8Value hostname(args[0]);
 
   int fam = AF_UNSPEC;
   if (args[1]->IsInt32()) {
@@ -736,8 +946,20 @@ static void Initialize(Handle<Object> target) {
   assert(r == ARES_SUCCESS);
 
   struct ares_options options;
-  uv_ares_init_options(uv_default_loop(), &ares_channel, &options, 0);
-  assert(r == 0);
+  memset(&options, 0, sizeof(options));
+  options.flags = ARES_FLAG_NOCHECKRESP;
+  options.sock_state_cb = ares_sockstate_cb;
+  options.sock_state_cb_data = uv_default_loop();
+
+  /* We do the call to ares_init_option for caller. */
+  r = ares_init_options(&ares_channel,
+                        &options,
+                        ARES_OPT_FLAGS | ARES_OPT_SOCK_STATE_CB);
+  assert(r == ARES_SUCCESS);
+
+  /* Initialize the timeout timer. The timer won't be started until the */
+  /* first socket is opened. */
+  uv_timer_init(uv_default_loop(), &ares_timer);
 
   NODE_SET_METHOD(target, "queryA", Query<QueryAWrap>);
   NODE_SET_METHOD(target, "queryAaaa", Query<QueryAaaaWrap>);
@@ -746,16 +968,21 @@ static void Initialize(Handle<Object> target) {
   NODE_SET_METHOD(target, "queryNs", Query<QueryNsWrap>);
   NODE_SET_METHOD(target, "queryTxt", Query<QueryTxtWrap>);
   NODE_SET_METHOD(target, "querySrv", Query<QuerySrvWrap>);
+  NODE_SET_METHOD(target, "queryNaptr", Query<QueryNaptrWrap>);
   NODE_SET_METHOD(target, "getHostByAddr", Query<GetHostByAddrWrap>);
   NODE_SET_METHOD(target, "getHostByName", QueryWithFamily<GetHostByNameWrap>);
 
   NODE_SET_METHOD(target, "getaddrinfo", GetAddrInfo);
+  NODE_SET_METHOD(target, "isIP", IsIP);
 
-  target->Set(String::NewSymbol("AF_INET"), Integer::New(AF_INET));
-  target->Set(String::NewSymbol("AF_INET6"), Integer::New(AF_INET6));
-  target->Set(String::NewSymbol("AF_UNSPEC"), Integer::New(AF_UNSPEC));
+  target->Set(String::NewSymbol("AF_INET"),
+              Integer::New(AF_INET));
+  target->Set(String::NewSymbol("AF_INET6"),
+              Integer::New(AF_INET6));
+  target->Set(String::NewSymbol("AF_UNSPEC"),
+              Integer::New(AF_UNSPEC));
 
-  oncomplete_sym = Persistent<String>::New(String::NewSymbol("oncomplete"));
+  oncomplete_sym = NODE_PSYMBOL("oncomplete");
 }
 
 
