@@ -19,11 +19,23 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-#include <node_dtrace.h>
-#include <string.h>
 
 #ifdef HAVE_DTRACE
+#include "node_dtrace.h"
+#include <string.h>
 #include "node_provider.h"
+#elif HAVE_ETW
+#include "node_dtrace.h"
+#include <string.h>
+#include "node_win32_etw_provider.h"
+#include "node_win32_etw_provider-inl.h"
+#elif HAVE_SYSTEMTAP
+#include <string.h>
+#include <node.h>
+#include <v8.h>
+#include <sys/sdt.h>
+#include "node_systemtap.h"
+#include "node_dtrace.h"
 #else
 #define NODE_HTTP_SERVER_REQUEST(arg0, arg1)
 #define NODE_HTTP_SERVER_REQUEST_ENABLED() (0)
@@ -54,7 +66,7 @@ using namespace v8;
     return (ThrowException(Exception::Error(String::New("expected " \
       "object for " #obj " to contain string member " #member)))); \
   } \
-  String::Utf8Value _##member(obj->Get(String::New(#member))->ToString()); \
+  String::Utf8Value _##member(obj->Get(String::New(#member))); \
   if ((*(const char **)valp = *_##member) == NULL) \
     *(const char **)valp = "<unknown>";
 
@@ -79,7 +91,12 @@ using namespace v8;
   } \
   node_dtrace_connection_t conn; \
   Local<Object> _##conn = Local<Object>::Cast(arg); \
-  SLURP_INT(_##conn, fd, &conn.fd); \
+  Local<Value> _handle = (_##conn)->Get(String::New("_handle")); \
+  if (_handle->IsObject()) { \
+    SLURP_INT(_handle.As<Object>(), fd, &conn.fd); \
+  } else { \
+    conn.fd = -1; \
+  } \
   SLURP_STRING(_##conn, remoteAddress, &conn.remote); \
   SLURP_INT(_##conn, remotePort, &conn.port); \
   SLURP_INT(_##conn, bufferSize, &conn.buffered);
@@ -115,67 +132,86 @@ using namespace v8;
 
 
 Handle<Value> DTRACE_NET_SERVER_CONNECTION(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_NET_SERVER_CONNECTION_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
 
   SLURP_CONNECTION(args[0], conn);
-  NODE_NET_SERVER_CONNECTION(&conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_NET_SERVER_CONNECTION(conn.fd, conn.remote, conn.port, \
+                             conn.buffered);
+#else
+  NODE_NET_SERVER_CONNECTION(&conn, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
 
 Handle<Value> DTRACE_NET_STREAM_END(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_NET_STREAM_END_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
 
   SLURP_CONNECTION(args[0], conn);
-  NODE_NET_STREAM_END(&conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_NET_STREAM_END(conn.fd, conn.remote, conn.port, conn.buffered);
+#else
+  NODE_NET_STREAM_END(&conn, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
 
 Handle<Value> DTRACE_NET_SOCKET_READ(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_NET_SOCKET_READ_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
-  int nbytes;
 
   SLURP_CONNECTION(args[0], conn);
 
+#ifdef HAVE_SYSTEMTAP
+  NODE_NET_SOCKET_READ(conn.fd, conn.remote, conn.port, conn.buffered);
+#else
   if (!args[1]->IsNumber()) {
     return (ThrowException(Exception::Error(String::New("expected " 
       "argument 1 to be number of bytes"))));
   }
-
-  nbytes = args[1]->Int32Value();
-
-  NODE_NET_SOCKET_READ(&conn, nbytes);
+  int nbytes = args[1]->Int32Value();
+  NODE_NET_SOCKET_READ(&conn, nbytes, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
 
 Handle<Value> DTRACE_NET_SOCKET_WRITE(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_NET_SOCKET_WRITE_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
-  int nbytes;
 
   SLURP_CONNECTION(args[0], conn);
 
+#ifdef HAVE_SYSTEMTAP
+  NODE_NET_SOCKET_WRITE(conn.fd, conn.remote, conn.port, conn.buffered);
+#else
   if (!args[1]->IsNumber()) {
     return (ThrowException(Exception::Error(String::New("expected " 
       "argument 1 to be number of bytes"))));
   }
-
-  nbytes = args[1]->Int32Value();
-
-  NODE_NET_SOCKET_WRITE(&conn, nbytes);
+  int nbytes = args[1]->Int32Value();
+  NODE_NET_SOCKET_WRITE(&conn, nbytes, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
@@ -183,8 +219,10 @@ Handle<Value> DTRACE_NET_SOCKET_WRITE(const Arguments& args) {
 Handle<Value> DTRACE_HTTP_SERVER_REQUEST(const Arguments& args) {
   node_dtrace_http_server_request_t req;
 
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_HTTP_SERVER_REQUEST_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
 
@@ -203,25 +241,37 @@ Handle<Value> DTRACE_HTTP_SERVER_REQUEST(const Arguments& args) {
       "object for request to contain string member headers"))));
 
   Local<Value> strfwdfor = headers->Get(String::New("x-forwarded-for"));
-  String::Utf8Value fwdfor(strfwdfor->ToString());
+  String::Utf8Value fwdfor(strfwdfor);
 
   if (!strfwdfor->IsString() || (req.forwardedFor = *fwdfor) == NULL)
     req.forwardedFor = const_cast<char*>("");
 
   SLURP_CONNECTION(args[1], conn);
 
-  NODE_HTTP_SERVER_REQUEST(&req, &conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_HTTP_SERVER_REQUEST(&req, conn.fd, conn.remote, conn.port, \
+                           conn.buffered);
+#else
+  NODE_HTTP_SERVER_REQUEST(&req, &conn, conn.remote, conn.port, req.method, \
+                           req.url, conn.fd);
+#endif
   return Undefined();
 }
 
 Handle<Value> DTRACE_HTTP_SERVER_RESPONSE(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_HTTP_SERVER_RESPONSE_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
 
   SLURP_CONNECTION(args[0], conn);
-  NODE_HTTP_SERVER_RESPONSE(&conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_HTTP_SERVER_RESPONSE(conn.fd, conn.remote, conn.port, conn.buffered);
+#else
+  NODE_HTTP_SERVER_RESPONSE(&conn, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
@@ -230,8 +280,10 @@ Handle<Value> DTRACE_HTTP_CLIENT_REQUEST(const Arguments& args) {
   node_dtrace_http_client_request_t req;
   char *header;
 
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_HTTP_CLIENT_REQUEST_ENABLED())
     return Undefined();
+#endif
 
   HandleScope scope;
 
@@ -260,26 +312,41 @@ Handle<Value> DTRACE_HTTP_CLIENT_REQUEST(const Arguments& args) {
   *header = '\0';
 
   SLURP_CONNECTION_HTTP_CLIENT(args[1], conn);
-  NODE_HTTP_CLIENT_REQUEST(&req, &conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_HTTP_CLIENT_REQUEST(&req, conn.fd, conn.remote, conn.port, \
+                           conn.buffered);
+#else
+  NODE_HTTP_CLIENT_REQUEST(&req, &conn, conn.remote, conn.port, req.method, \
+                           req.url, conn.fd);
+#endif
   return Undefined();
 }
 
 Handle<Value> DTRACE_HTTP_CLIENT_RESPONSE(const Arguments& args) {
+#ifndef HAVE_SYSTEMTAP
   if (!NODE_HTTP_CLIENT_RESPONSE_ENABLED())
     return Undefined();
-
+#endif
   HandleScope scope;
 
   SLURP_CONNECTION_HTTP_CLIENT_RESPONSE(args[0], args[1], conn);
-  NODE_HTTP_CLIENT_RESPONSE(&conn);
+#ifdef HAVE_SYSTEMTAP
+  NODE_HTTP_CLIENT_RESPONSE(conn.fd, conn.remote, conn.port, conn.buffered);
+#else
+  NODE_HTTP_CLIENT_RESPONSE(&conn, conn.remote, conn.port, conn.fd);
+#endif
 
   return Undefined();
 }
 
-#define NODE_PROBE(name) #name, name
+#define NODE_PROBE(name) #name, name, Persistent<FunctionTemplate>()
 
 static int dtrace_gc_start(GCType type, GCCallbackFlags flags) {
+#ifdef HAVE_SYSTEMTAP
+  NODE_GC_START();
+#else
   NODE_GC_START(type, flags);
+#endif
   /*
    * We avoid the tail-call elimination of the USDT probe (which screws up
    * args) by forcing a return of 0.
@@ -288,7 +355,11 @@ static int dtrace_gc_start(GCType type, GCCallbackFlags flags) {
 }
 
 static int dtrace_gc_done(GCType type, GCCallbackFlags flags) {
+#ifdef HAVE_SYSTEMTAP
+  NODE_GC_DONE();
+#else
   NODE_GC_DONE(type, flags);
+#endif
   return 0;
 }
 
@@ -305,17 +376,20 @@ void InitDTrace(Handle<Object> target) {
     { NODE_PROBE(DTRACE_HTTP_SERVER_REQUEST) },
     { NODE_PROBE(DTRACE_HTTP_SERVER_RESPONSE) },
     { NODE_PROBE(DTRACE_HTTP_CLIENT_REQUEST) },
-    { NODE_PROBE(DTRACE_HTTP_CLIENT_RESPONSE) },
-    { NULL }
+    { NODE_PROBE(DTRACE_HTTP_CLIENT_RESPONSE) }
   };
 
-  for (int i = 0; tab[i].name != NULL; i++) {
+  for (unsigned int i = 0; i < ARRAY_SIZE(tab); i++) {
     tab[i].templ = Persistent<FunctionTemplate>::New(
         FunctionTemplate::New(tab[i].func));
     target->Set(String::NewSymbol(tab[i].name), tab[i].templ->GetFunction());
   }
 
-#ifdef HAVE_DTRACE
+#ifdef HAVE_ETW
+  init_etw();
+#endif
+
+#if defined HAVE_DTRACE || defined HAVE_ETW || defined HAVE_SYSTEMTAP
   v8::V8::AddGCPrologueCallback((GCPrologueCallback)dtrace_gc_start);
   v8::V8::AddGCEpilogueCallback((GCEpilogueCallback)dtrace_gc_done);
 #endif
