@@ -19,13 +19,11 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-#include <node_http_parser.h>
+#include "node_http_parser.h"
 
-#include <v8.h>
-#include <node.h>
-#include <node_buffer.h>
-
-#include <http_parser.h>
+#include "v8.h"
+#include "node.h"
+#include "node_buffer.h"
 
 #include <string.h>  /* strdup() */
 #if !defined(_MSC_VER)
@@ -56,32 +54,6 @@ static Persistent<String> on_headers_complete_sym;
 static Persistent<String> on_body_sym;
 static Persistent<String> on_message_complete_sym;
 
-static Persistent<String> delete_sym;
-static Persistent<String> get_sym;
-static Persistent<String> head_sym;
-static Persistent<String> post_sym;
-static Persistent<String> put_sym;
-static Persistent<String> connect_sym;
-static Persistent<String> options_sym;
-static Persistent<String> trace_sym;
-static Persistent<String> patch_sym;
-static Persistent<String> copy_sym;
-static Persistent<String> lock_sym;
-static Persistent<String> mkcol_sym;
-static Persistent<String> move_sym;
-static Persistent<String> propfind_sym;
-static Persistent<String> proppatch_sym;
-static Persistent<String> unlock_sym;
-static Persistent<String> report_sym;
-static Persistent<String> mkactivity_sym;
-static Persistent<String> checkout_sym;
-static Persistent<String> merge_sym;
-static Persistent<String> msearch_sym;
-static Persistent<String> notify_sym;
-static Persistent<String> subscribe_sym;
-static Persistent<String> unsubscribe_sym;
-static Persistent<String> unknown_method_sym;
-
 static Persistent<String> method_sym;
 static Persistent<String> status_code_sym;
 static Persistent<String> http_version_sym;
@@ -92,6 +64,12 @@ static Persistent<String> upgrade_sym;
 static Persistent<String> headers_sym;
 static Persistent<String> url_sym;
 
+static Persistent<String> unknown_method_sym;
+
+#define X(num, name, string) static Persistent<String> name##_sym;
+HTTP_METHOD_MAP(X)
+#undef X
+
 static struct http_parser_settings settings;
 
 
@@ -103,51 +81,30 @@ static char* current_buffer_data;
 static size_t current_buffer_len;
 
 
-#define HTTP_CB(name)                                               \
-	  static int name(http_parser* p_) {                              \
-	    Parser* self = container_of(p_, Parser, parser_);             \
-	    return self->name##_();                                       \
-	  }                                                               \
-	  int name##_()
+#define HTTP_CB(name)                                                         \
+  static int name(http_parser* p_) {                                          \
+    Parser* self = container_of(p_, Parser, parser_);                         \
+    return self->name##_();                                                   \
+  }                                                                           \
+  int name##_()
 
 
-#define HTTP_DATA_CB(name)                                          \
-  static int name(http_parser* p_, const char* at, size_t length) { \
-    Parser* self = container_of(p_, Parser, parser_);               \
-    return self->name##_(at, length);                               \
-  }                                                                 \
+#define HTTP_DATA_CB(name)                                                    \
+  static int name(http_parser* p_, const char* at, size_t length) {           \
+    Parser* self = container_of(p_, Parser, parser_);                         \
+    return self->name##_(at, length);                                         \
+  }                                                                           \
   int name##_(const char* at, size_t length)
 
 
 static inline Persistent<String>
 method_to_str(unsigned short m) {
   switch (m) {
-    case HTTP_DELETE:     return delete_sym;
-    case HTTP_GET:        return get_sym;
-    case HTTP_HEAD:       return head_sym;
-    case HTTP_POST:       return post_sym;
-    case HTTP_PUT:        return put_sym;
-    case HTTP_CONNECT:    return connect_sym;
-    case HTTP_OPTIONS:    return options_sym;
-    case HTTP_TRACE:      return trace_sym;
-    case HTTP_PATCH:      return patch_sym;
-    case HTTP_COPY:       return copy_sym;
-    case HTTP_LOCK:       return lock_sym;
-    case HTTP_MKCOL:      return mkcol_sym;
-    case HTTP_MOVE:       return move_sym;
-    case HTTP_PROPFIND:   return propfind_sym;
-    case HTTP_PROPPATCH:  return proppatch_sym;
-    case HTTP_UNLOCK:     return unlock_sym;
-    case HTTP_REPORT:     return report_sym;
-    case HTTP_MKACTIVITY: return mkactivity_sym;
-    case HTTP_CHECKOUT:   return checkout_sym;
-    case HTTP_MERGE:      return merge_sym;
-    case HTTP_MSEARCH:    return msearch_sym;
-    case HTTP_NOTIFY:     return notify_sym;
-    case HTTP_SUBSCRIBE:  return subscribe_sym;
-    case HTTP_UNSUBSCRIBE:return unsubscribe_sym;
-    default:              return unknown_method_sym;
+#define X(num, name, string) case HTTP_##name: return name##_sym;
+  HTTP_METHOD_MAP(X)
+#undef X
   }
+  return unknown_method_sym;
 }
 
 
@@ -209,7 +166,7 @@ struct StringPtr {
   }
 
 
-  Handle<String> ToString() const {
+  Local<String> ToString() const {
     if (str_)
       return String::New(str_, size_);
     else
@@ -312,17 +269,23 @@ public:
 
     // STATUS
     if (parser_.type == HTTP_RESPONSE) {
-      message_info->Set(status_code_sym, Integer::New(parser_.status_code));
+      message_info->Set(status_code_sym,
+                        Integer::New(parser_.status_code));
     }
 
     // VERSION
-    message_info->Set(version_major_sym, Integer::New(parser_.http_major));
-    message_info->Set(version_minor_sym, Integer::New(parser_.http_minor));
+    message_info->Set(version_major_sym,
+                      Integer::New(parser_.http_major));
+    message_info->Set(version_minor_sym,
+                      Integer::New(parser_.http_minor));
 
     message_info->Set(should_keep_alive_sym,
-        http_should_keep_alive(&parser_) ? True() : False());
+                      http_should_keep_alive(&parser_) ? True()
+                                                       : False());
 
-    message_info->Set(upgrade_sym, parser_.upgrade ? True() : False());
+    message_info->Set(upgrade_sym,
+                      parser_.upgrade ? True()
+                                      : False());
 
     Local<Value> argv[1] = { message_info };
 
@@ -345,7 +308,7 @@ public:
     if (!cb->IsFunction())
       return 0;
 
-    Handle<Value> argv[3] = {
+    Local<Value> argv[3] = {
       *current_buffer,
       Integer::New(at - current_buffer_data),
       Integer::New(length)
@@ -449,7 +412,7 @@ public:
     size_t len = args[2]->Int32Value();
     if (off+len > buffer_len) {
       return ThrowException(Exception::Error(
-            String::New("Length is extends beyond buffer")));
+            String::New("off + len > buffer.length")));
     }
 
     // Assign 'buffer_' while we parse. The callbacks will access that varible.
@@ -475,9 +438,12 @@ public:
     // If there was a parse error in one of the callbacks
     // TODO What if there is an error on EOF?
     if (!parser->parser_.upgrade && nparsed != len) {
+      enum http_errno err = HTTP_PARSER_ERRNO(&parser->parser_);
+
       Local<Value> e = Exception::Error(String::NewSymbol("Parse Error"));
       Local<Object> obj = e->ToObject();
       obj->Set(String::NewSymbol("bytesParsed"), nparsed_obj);
+      obj->Set(String::NewSymbol("code"), String::New(http_errno_name(err)));
       return scope.Close(e);
     } else {
       return scope.Close(nparsed_obj);
@@ -498,9 +464,12 @@ public:
     if (parser->got_exception_) return Local<Value>();
 
     if (rv != 0) {
+      enum http_errno err = HTTP_PARSER_ERRNO(&parser->parser_);
+
       Local<Value> e = Exception::Error(String::NewSymbol("Parse Error"));
       Local<Object> obj = e->ToObject();
       obj->Set(String::NewSymbol("bytesParsed"), Integer::New(0));
+      obj->Set(String::NewSymbol("code"), String::New(http_errno_name(err)));
       return scope.Close(e);
     }
 
@@ -526,6 +495,15 @@ public:
   }
 
 
+  template <bool should_pause>
+  static Handle<Value> Pause(const Arguments& args) {
+    HandleScope scope;
+    Parser* parser = ObjectWrap::Unwrap<Parser>(args.This());
+    http_parser_pause(&parser->parser_, should_pause);
+    return Undefined();
+  }
+
+
 private:
 
   Local<Array> CreateHeaders() {
@@ -551,7 +529,7 @@ private:
     if (!cb->IsFunction())
       return;
 
-    Handle<Value> argv[2] = {
+    Local<Value> argv[2] = {
       CreateHeaders(),
       url_.ToString()
     };
@@ -595,12 +573,18 @@ void InitHttpParser(Handle<Object> target) {
   t->SetClassName(String::NewSymbol("HTTPParser"));
 
   PropertyAttribute attrib = (PropertyAttribute) (ReadOnly | DontDelete);
-  t->Set(String::NewSymbol("REQUEST"), Integer::New(HTTP_REQUEST), attrib);
-  t->Set(String::NewSymbol("RESPONSE"), Integer::New(HTTP_RESPONSE), attrib);
+  t->Set(String::NewSymbol("REQUEST"),
+         Integer::New(HTTP_REQUEST),
+         attrib);
+  t->Set(String::NewSymbol("RESPONSE"),
+         Integer::New(HTTP_RESPONSE),
+         attrib);
 
   NODE_SET_PROTOTYPE_METHOD(t, "execute", Parser::Execute);
   NODE_SET_PROTOTYPE_METHOD(t, "finish", Parser::Finish);
   NODE_SET_PROTOTYPE_METHOD(t, "reinitialize", Parser::Reinitialize);
+  NODE_SET_PROTOTYPE_METHOD(t, "pause", Parser::Pause<true>);
+  NODE_SET_PROTOTYPE_METHOD(t, "resume", Parser::Pause<false>);
 
   target->Set(String::NewSymbol("HTTPParser"), t->GetFunction());
 
@@ -609,30 +593,9 @@ void InitHttpParser(Handle<Object> target) {
   on_body_sym             = NODE_PSYMBOL("onBody");
   on_message_complete_sym = NODE_PSYMBOL("onMessageComplete");
 
-  delete_sym = NODE_PSYMBOL("DELETE");
-  get_sym = NODE_PSYMBOL("GET");
-  head_sym = NODE_PSYMBOL("HEAD");
-  post_sym = NODE_PSYMBOL("POST");
-  put_sym = NODE_PSYMBOL("PUT");
-  connect_sym = NODE_PSYMBOL("CONNECT");
-  options_sym = NODE_PSYMBOL("OPTIONS");
-  trace_sym = NODE_PSYMBOL("TRACE");
-  patch_sym = NODE_PSYMBOL("PATCH");
-  copy_sym = NODE_PSYMBOL("COPY");
-  lock_sym = NODE_PSYMBOL("LOCK");
-  mkcol_sym = NODE_PSYMBOL("MKCOL");
-  move_sym = NODE_PSYMBOL("MOVE");
-  propfind_sym = NODE_PSYMBOL("PROPFIND");
-  proppatch_sym = NODE_PSYMBOL("PROPPATCH");
-  unlock_sym = NODE_PSYMBOL("UNLOCK");
-  report_sym = NODE_PSYMBOL("REPORT");
-  mkactivity_sym = NODE_PSYMBOL("MKACTIVITY");
-  checkout_sym = NODE_PSYMBOL("CHECKOUT");
-  merge_sym = NODE_PSYMBOL("MERGE");
-  msearch_sym = NODE_PSYMBOL("M-SEARCH");
-  notify_sym = NODE_PSYMBOL("NOTIFY");
-  subscribe_sym = NODE_PSYMBOL("SUBSCRIBE");
-  unsubscribe_sym = NODE_PSYMBOL("UNSUBSCRIBE");;
+#define X(num, name, string) name##_sym = NODE_PSYMBOL(#string);
+  HTTP_METHOD_MAP(X)
+#undef X
   unknown_method_sym = NODE_PSYMBOL("UNKNOWN_METHOD");
 
   method_sym = NODE_PSYMBOL("method");
