@@ -19,7 +19,7 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-#include <node_stat_watcher.h>
+#include "node_stat_watcher.h"
 
 #include <assert.h>
 #include <string.h>
@@ -30,22 +30,18 @@ namespace node {
 using namespace v8;
 
 Persistent<FunctionTemplate> StatWatcher::constructor_template;
+static Persistent<String> onchange_sym;
+static Persistent<String> onstop_sym;
 
-static Persistent<String> change_symbol;
-static Persistent<String> stop_symbol;
 
 void StatWatcher::Initialize(Handle<Object> target) {
   HandleScope scope;
 
   Local<FunctionTemplate> t = FunctionTemplate::New(StatWatcher::New);
   constructor_template = Persistent<FunctionTemplate>::New(t);
-  constructor_template->Inherit(EventEmitter::constructor_template);
   constructor_template->InstanceTemplate()->SetInternalFieldCount(1);
   constructor_template->SetClassName(String::NewSymbol("StatWatcher"));
 
-  change_symbol = NODE_PSYMBOL("change");
-  stop_symbol = NODE_PSYMBOL("stop");
-
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "start", StatWatcher::Start);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "stop", StatWatcher::Stop);
 
@@ -53,58 +49,68 @@ void StatWatcher::Initialize(Handle<Object> target) {
 }
 
 
-void StatWatcher::Callback(EV_P_ ev_stat *watcher, int revents) {
-  assert(revents == EV_STAT);
-  StatWatcher *handler = static_cast<StatWatcher*>(watcher->data);
-  assert(watcher == &handler->watcher_);
+static void Delete(uv_handle_t* handle) {
+  delete reinterpret_cast<uv_fs_poll_t*>(handle);
+}
+
+
+StatWatcher::StatWatcher()
+  : ObjectWrap()
+  , watcher_(new uv_fs_poll_t)
+{
+  uv_fs_poll_init(uv_default_loop(), watcher_);
+  watcher_->data = static_cast<void*>(this);
+}
+
+
+StatWatcher::~StatWatcher() {
+  Stop();
+  uv_close(reinterpret_cast<uv_handle_t*>(watcher_), Delete);
+}
+
+
+void StatWatcher::Callback(uv_fs_poll_t* handle,
+                           int status,
+                           const uv_statbuf_t* prev,
+                           const uv_statbuf_t* curr) {
+  StatWatcher* wrap = static_cast<StatWatcher*>(handle->data);
+  assert(wrap->watcher_ == handle);
   HandleScope scope;
-  Handle<Value> argv[2];
-  argv[0] = Handle<Value>(BuildStatsObject(&watcher->attr));
-  argv[1] = Handle<Value>(BuildStatsObject(&watcher->prev));
-  handler->Emit(change_symbol, 2, argv);
+  Local<Value> argv[3];
+  argv[0] = BuildStatsObject(curr);
+  argv[1] = BuildStatsObject(prev);
+  argv[2] = Integer::New(status);
+  if (status == -1) {
+    SetErrno(uv_last_error(wrap->watcher_->loop));
+  }
+  if (onchange_sym.IsEmpty()) {
+    onchange_sym = NODE_PSYMBOL("onchange");
+  }
+  MakeCallback(wrap->handle_, onchange_sym, ARRAY_SIZE(argv), argv);
 }
 
 
 Handle<Value> StatWatcher::New(const Arguments& args) {
-  if (!args.IsConstructCall()) {
-    return FromConstructorTemplate(constructor_template, args);
-  }
-
+  assert(args.IsConstructCall());
   HandleScope scope;
-  StatWatcher *s = new StatWatcher();
+  StatWatcher* s = new StatWatcher();
   s->Wrap(args.Holder());
   return args.This();
 }
 
 
 Handle<Value> StatWatcher::Start(const Arguments& args) {
+  assert(args.Length() == 3);
   HandleScope scope;
 
-  if (args.Length() < 1 || !args[0]->IsString()) {
-    return ThrowException(Exception::TypeError(String::New("Bad arguments")));
-  }
+  StatWatcher* wrap = ObjectWrap::Unwrap<StatWatcher>(args.Holder());
+  String::Utf8Value path(args[0]);
+  const bool persistent = args[1]->BooleanValue();
+  const uint32_t interval = args[2]->Uint32Value();
 
-  StatWatcher *handler = ObjectWrap::Unwrap<StatWatcher>(args.Holder());
-  String::Utf8Value path(args[0]->ToString());
-
-  assert(handler->path_ == NULL);
-  handler->path_ = strdup(*path);
-
-  ev_tstamp interval = 0.;
-  if (args[2]->IsInt32()) {
-    interval = NODE_V8_UNIXTIME(args[2]);
-  }
-
-  ev_stat_set(&handler->watcher_, handler->path_, interval);
-  ev_stat_start(EV_DEFAULT_UC_ &handler->watcher_);
-
-  handler->persistent_ = args[1]->IsTrue();
-
-  if (!handler->persistent_) {
-    ev_unref(EV_DEFAULT_UC);
-  }
-
-  handler->Ref();
+  if (!persistent) uv_unref(reinterpret_cast<uv_handle_t*>(wrap->watcher_));
+  uv_fs_poll_start(wrap->watcher_, Callback, *path, interval);
+  wrap->Ref();
 
   return Undefined();
 }
@@ -112,21 +118,20 @@ Handle<Value> StatWatcher::Start(const Arguments& args) {
 
 Handle<Value> StatWatcher::Stop(const Arguments& args) {
   HandleScope scope;
-  StatWatcher *handler = ObjectWrap::Unwrap<StatWatcher>(args.Holder());
-  handler->Emit(stop_symbol, 0, NULL);
-  handler->Stop();
+  StatWatcher* wrap = ObjectWrap::Unwrap<StatWatcher>(args.Holder());
+  if (onstop_sym.IsEmpty()) {
+    onstop_sym = NODE_PSYMBOL("onstop");
+  }
+  MakeCallback(wrap->handle_, onstop_sym, 0, NULL);
+  wrap->Stop();
   return Undefined();
 }
 
 
 void StatWatcher::Stop () {
-  if (watcher_.active) {
-    if (!persistent_) ev_ref(EV_DEFAULT_UC);
-    ev_stat_stop(EV_DEFAULT_UC_ &watcher_);
-    free(path_);
-    path_ = NULL;
-    Unref();
-  }
+  if (!uv_is_active(reinterpret_cast<uv_handle_t*>(watcher_))) return;
+  uv_fs_poll_stop(watcher_);
+  Unref();
 }
 
 
