@@ -26,10 +26,6 @@ common.print('hello world\r\n');
 
 var stdin = process.openStdin();
 
-stdin.addListener('data', function(data) {
+stdin.on('data', function(data) {
   process.stdout.write(data.toString());
 });
-
-stdin.addListener('end', function() {
-  process.stdout.end();
-});
