@@ -19,6 +19,9 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+
+
+
 // This test requires the program 'ab'
 var common = require('../common');
 var assert = require('assert');
@@ -80,7 +83,7 @@ server.listen(common.PORT, function() {
   });
 });
 
-process.addListener('exit', function() {
+process.on('exit', function() {
   assert.equal(true, normalReqSec > 50);
   assert.equal(true, keepAliveReqSec > 50);
   assert.equal(true, normalReqSec < keepAliveReqSec);
