@@ -20,13 +20,6 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 var common = require('../common');
-var assert = require('assert');
-var fork = require('child_process').fork;
+var child_process = require('child_process');
 
-var filename = common.fixturesDir + '/destroy-stdin.js';
-
-// Ensure that we don't accidentally close fd 0 and
-// reuse it for something else, it causes all kinds
-// of obscure bugs.
-process.stdin.destroy();
-fork(filename).stdin.on('end', process.exit);
+child_process.fork(common.fixturesDir + '/empty.js'); // should not hang
