@@ -19,45 +19,50 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-var common = require('../common.js'),
-    assert = require('assert'),
-    child = require('child_process'),
-    nodejs = '"' + process.execPath + '"';
-
-// replace \ by / because windows uses backslashes in paths, but they're still
-// interpreted as the escape character when put between quotes.
-var filename = __filename.replace(/\\/g, '/');
-
 if (module.parent) {
   // signal we've been loaded as a module
   console.log('Loaded as a module, exiting with status code 42.');
   process.exit(42);
 }
 
+var common = require('../common.js'),
+    assert = require('assert'),
+    child = require('child_process'),
+    nodejs = '"' + process.execPath + '"';
+
+
+// replace \ by / because windows uses backslashes in paths, but they're still
+// interpreted as the escape character when put between quotes.
+var filename = __filename.replace(/\\/g, '/');
+
 // assert that nothing is written to stdout
 child.exec(nodejs + ' --eval 42',
     function(err, stdout, stderr) {
       assert.equal(stdout, '');
+      assert.equal(stderr, '');
     });
 
 // assert that "42\n" is written to stderr
 child.exec(nodejs + ' --eval "console.error(42)"',
     function(err, stdout, stderr) {
+      assert.equal(stdout, '');
       assert.equal(stderr, '42\n');
     });
 
-// assert that nothing is written to stdout
-['--print --eval', '-p -e', '-pe'].forEach(function(s) {
+// assert that the expected output is written to stdout
+['--print', '-p -e', '-pe', '-p'].forEach(function(s) {
   var cmd = nodejs + ' ' + s + ' ';
 
   child.exec(cmd + '42',
       function(err, stdout, stderr) {
         assert.equal(stdout, '42\n');
+        assert.equal(stderr, '');
       });
 
   child.exec(cmd + "'[]'",
       function(err, stdout, stderr) {
         assert.equal(stdout, '[]\n');
+        assert.equal(stderr, '');
       });
 });
 
@@ -78,3 +83,10 @@ child.exec(nodejs + ' -e ""', function(status, stdout, stderr) {
   assert.equal(stdout, '');
   assert.equal(stderr, '');
 });
+
+// "\\-42" should be interpreted as an escaped expression, not a switch
+child.exec(nodejs + ' -p "\\-42"',
+    function(err, stdout, stderr) {
+      assert.equal(stdout, '-42\n');
+      assert.equal(stderr, '');
+    });
