@@ -40,10 +40,8 @@ var server = net.createServer(function(conn) {
       '\r\n' +
       body;
 
-  conn.write(response, function() {
-    conn.destroy();
-    server.close();
-  });
+  conn.end(response);
+  server.close();
 });
 
 server.listen(common.PORT, function() {
@@ -51,6 +49,7 @@ server.listen(common.PORT, function() {
     assert.equal(res.headers['content-type'],
                  'text/plain;x-unix-mode=0600;name="hello.txt"');
     gotResponse = true;
+    res.destroy();
   });
 });
 
