@@ -30,8 +30,12 @@ var http = require('http');
 var srv = http.createServer(function(req, res) {
   assert.equal(req.headers.accept, 'abc, def, ghijklmnopqrst');
   assert.equal(req.headers.host, 'foo');
+  assert.equal(req.headers['www-authenticate'], 'foo, bar, baz');
+  assert.equal(req.headers['proxy-authenticate'], 'foo, bar, baz');
   assert.equal(req.headers['x-foo'], 'bingo');
   assert.equal(req.headers['x-bar'], 'banjo, bango');
+  assert.equal(req.headers['sec-websocket-protocol'], 'chat, share');
+  assert.equal(req.headers['sec-websocket-extensions'], 'foo; 1, bar; 2, baz');
 
   res.writeHead(200, {'Content-Type' : 'text/plain'});
   res.end('EOF');
@@ -51,9 +55,20 @@ srv.listen(common.PORT, function() {
       ['host', 'foo'],
       ['Host', 'bar'],
       ['hOst', 'baz'],
+      ['www-authenticate', 'foo'],
+      ['WWW-Authenticate', 'bar'],
+      ['WWW-AUTHENTICATE', 'baz'],
+      ['proxy-authenticate','foo'],
+      ['Proxy-Authenticate','bar'],
+      ['PROXY-AUTHENTICATE','baz'],
       ['x-foo', 'bingo'],
       ['x-bar', 'banjo'],
-      ['x-bar', 'bango']
+      ['x-bar', 'bango'],
+      ['sec-websocket-protocol', 'chat'],
+      ['sec-websocket-protocol', 'share'],
+      ['sec-websocket-extensions', 'foo; 1'],
+      ['sec-websocket-extensions', 'bar; 2'],
+      ['sec-websocket-extensions', 'baz']
     ]
   });
 });
