@@ -30,39 +30,47 @@ var client_recv_count = 0;
 var disconnect_count = 0;
 
 var server = net.createServer(function(socket) {
-  socket.on('connect', function() {
-    socket.write('hello\r\n');
-  });
+  console.error('SERVER: got socket connection');
+  socket.resume();
+
+  console.error('SERVER connect, writing');
+  socket.write('hello\r\n');
 
   socket.on('end', function() {
+    console.error('SERVER socket end, calling end()');
     socket.end();
   });
 
   socket.on('close', function(had_error) {
-    //console.log('server had_error: ' + JSON.stringify(had_error));
+    console.log('SERVER had_error: ' + JSON.stringify(had_error));
     assert.equal(false, had_error);
   });
 });
 
 server.listen(common.PORT, function() {
-  console.log('listening');
+  console.log('SERVER listening');
   var client = net.createConnection(common.PORT);
 
   client.setEncoding('UTF8');
 
   client.on('connect', function() {
-    console.log('client connected.');
+    console.error('CLIENT connected', client._writableState);
   });
 
   client.on('data', function(chunk) {
     client_recv_count += 1;
     console.log('client_recv_count ' + client_recv_count);
     assert.equal('hello\r\n', chunk);
+    console.error('CLIENT: calling end', client._writableState);
     client.end();
   });
 
+  client.on('end', function() {
+    console.error('CLIENT end');
+  });
+
   client.on('close', function(had_error) {
-    console.log('disconnect');
+    console.log('CLIENT disconnect');
     assert.equal(false, had_error);
     if (disconnect_count++ < N)
       client.connect(common.PORT); // reconnect
