@@ -19,6 +19,7 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+var common = require('../common');
 var assert = require('assert');
 var util = require('util');
 
@@ -29,7 +30,7 @@ function ArrayStream() {
   this.run = function(data) {
     var self = this;
     data.forEach(function(line) {
-      self.emit('data', line);
+      self.emit('data', line + '\n');
     });
   }
 }
@@ -54,6 +55,9 @@ putIn.run([
 testMe.complete('inner.o', function(error, data) {
   assert.deepEqual(data, doesNotBreak);
 });
+testMe.complete('console.lo', function(error, data) {
+  assert.deepEqual(data, [['console.log'], 'console.lo']);
+});
 
 // Tab Complete will return globaly scoped variables
 putIn.run(['};']);
@@ -107,7 +111,7 @@ testMe.complete('inner.o', function(error, data) {
 putIn.run(['.clear']);
 
 // Tab Complete will return a complex local variable even if the function
-// has paramaters
+// has parameters
 putIn.run([
   'var top = function (one, two) {',
   'var inner = {',
@@ -136,7 +140,7 @@ testMe.complete('inner.o', function(error, data) {
 putIn.run(['.clear']);
 
 // currently does not work, but should not break note the inner function
-// def has the params and { on a seperate line
+// def has the params and { on a separate line
 putIn.run([
   'var top = function () {',
   'r = function test (',
@@ -187,5 +191,23 @@ putIn.run([
   'var str = "test";'
 ]);
 testMe.complete('str.len', function(error, data) {
-  assert.deepEqual(data, [ [ 'str.length' ], 'str.len' ]);
+  assert.deepEqual(data, [['str.length'], 'str.len']);
+});
+
+putIn.run(['.clear']);
+
+// tab completion should not break on spaces
+var spaceTimeout = setTimeout(function() {
+  throw new Error('timeout');
+}, 1000);
+
+testMe.complete(' ', function(error, data) {
+  assert.deepEqual(data, [[],undefined]);
+  clearTimeout(spaceTimeout);
+});
+
+// tab completion should pick up the global "toString" object, and
+// any other properties up the "global" object's prototype chain
+testMe.complete('toSt', function(error, data) {
+  assert.deepEqual(data, [['toString'], 'toSt']);
 });
