@@ -42,7 +42,6 @@ result = script.runInContext(context);
 assert.equal(3, context.foo);
 assert.equal('lala', context.thing);
 
-
 // Issue GH-227:
 Script.runInNewContext('', null, 'some.js');
 
@@ -65,7 +64,14 @@ function isTypeError(o) {
   return o instanceof TypeError;
 }
 
-[undefined, null, 0, 0.0, '', {}, []].forEach(function(e) {
+([undefined, null, 0, 0.0, '', {}, []].forEach(function(e) {
   assert.throws(function() { script.runInContext(e); }, isTypeError);
   assert.throws(function() { vm.runInContext('', e); }, isTypeError);
-});
+}));
+
+// Issue GH-693:
+common.debug('test RegExp as argument to assert.throws');
+script = vm.createScript('var assert = require(\'assert\'); assert.throws(' +
+                         'function() { throw "hello world"; }, /hello/);',
+                         'some.js');
+script.runInNewContext({ require : require });
