@@ -19,6 +19,9 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+
+
+
 var common = require('../common');
 var assert = require('assert');
 var join = require('path').join;
@@ -28,7 +31,8 @@ var fs = require('fs');
 var stdoutScript = join(common.fixturesDir, 'echo.js');
 var tmpFile = join(common.fixturesDir, 'stdin.txt');
 
-var cmd = '"' + process.argv[0] + '" "' + stdoutScript + '" < "' + tmpFile + '"';
+var cmd = '"' + process.argv[0] + '" "' + stdoutScript + '" < "' +
+    tmpFile + '"';
 
 var string = 'abc\nümlaut.\nsomething else\n' +
              '南越国是前203年至前111年存在于岭南地区的一个国家，国都位于番禺，' +
