@@ -30,7 +30,7 @@ try {
 } catch (e) {
   exceptionCaught = true;
   assert.ok(common.isError(e));
-  assert.equal('process.stdout cannot be closed', e.message);
+  assert.equal('process.stdout cannot be closed.', e.message);
 }
 
 assert.ok(exceptionCaught);
